from math import sqrt

from primalscheme3.core.config import Config


def ol_pp_score(
    pp_r_start: int, pp_n_p: int, leading_edge: int, config: Config
) -> float:
    """
    Higher score is better
    """
    dist_extend = pp_r_start - config.min_overlap - leading_edge
    prop_extended = dist_extend / config.amplicon_size_max

    return prop_extended**2 / sqrt(pp_n_p)


def bt_ol_pp_score(
    pp_r_start: int, pp_n_p: int, leading_edge: int, config: Config
) -> float:
    """
    Higher score is better
    """
    dist_extend = pp_r_start - config.min_overlap - leading_edge
    prop_extended = dist_extend / config.amplicon_size_max

    return prop_extended**2 / (sqrt(pp_n_p) / 2)


def walk_pp_score(pp_f_end: int, pp_n_p: int, leading_edge: int) -> float:
    """Higher score is better"""

    # If this is this is positive the primerpair is before the leading edge (HALF COVERAGE!!!)
    dist_from_leading_edge = leading_edge - pp_f_end

    # If the primerpair is before the leading edge favor it massively
    if dist_from_leading_edge > 0:
        return dist_from_leading_edge / sqrt(pp_n_p)
    else:
        return dist_from_leading_edge * sqrt(pp_n_p)



from click import UsageError


class CustomErrors(Exception):
    """Base class for custom errors"""

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other)

    def __hash__(self) -> int:
        return hash(type(self))


class GapOnSetBase(CustomErrors):
    """
    Defining position contains a gap
    """

    pass


class ContainsInvalidBase(CustomErrors):
    """
    Contains an invalid base
    """

    pass


class WalksOut(CustomErrors):
    """Walks out of the index of the MSA"""

    pass


class WalksTooFar(CustomErrors):
    """Primer walks more than primer_max_walk"""

    pass


class CustomRecursionError(CustomErrors):
    """Walks out of the index of the MSA"""

    pass


ERROR_SET = {
    WalksOut(),
    CustomRecursionError(),
    ContainsInvalidBase(),
    CustomErrors(),
    GapOnSetBase(),
    WalksTooFar(),
}


# MSA file errors
class MSAFileInvalid(UsageError):
    """Error raised when the MSA file is invalid"""

    pass


class MSAFileInvalidBase(MSAFileInvalid):
    """Error raised when the MSA file contains invalid bases"""

    pass


class MSAFileInvalidLength(MSAFileInvalid):
    """Error raised when the MSA file contains sequences of different lengths"""

    pass


class MSAFileDuplicateID(MSAFileInvalid):
    """Error raised when the MSA file contains sequences of different lengths"""

    pass


# Bed file errors
class BEDFileInvalid(UsageError):
    """Error raised when the BED file is invalid"""

    pass


# Digestion Errors
class DigestionFail(UsageError):
    """Base class for digestion failures"""

    pass


class DigestionFailNoPrimerPairs(DigestionFail):
    """Error raised when no primer pairs are found"""

    pass



import hashlib
import json
import pathlib
from time import sleep

from Bio import Seq, SeqIO, SeqRecord
from click import UsageError

# Interaction checker
from primaldimer_py import do_pools_interact_py  # type: ignore

from primalscheme3.core.bedfiles import (
    read_in_bedprimerpairs,
    read_in_extra_primers,
)
from primalscheme3.core.config import Config, MappingType
from primalscheme3.core.create_report_data import (
    generate_all_plotdata,
)
from primalscheme3.core.create_reports import generate_all_plots_html
from primalscheme3.core.errors import DigestionFailNoPrimerPairs
from primalscheme3.core.logger import setup_rich_logger
from primalscheme3.core.mapping import (
    generate_consensus,
    generate_reference,
)
from primalscheme3.core.mismatches import MatchDB
from primalscheme3.core.msa import MSA
from primalscheme3.core.primer_visual import primer_mismatch_heatmap
from primalscheme3.core.progress_tracker import ProgressManager
from primalscheme3.scheme.classes import Scheme, SchemeReturn


def schemereplace(
    config_path: pathlib.Path,
    ampliconsizemax: int,
    ampliconsizemin: int,
    primerbed: pathlib.Path,
    primername: str,
    msapath: pathlib.Path,
    pm: ProgressManager | None,
):
    """
    List all replacements primers
    """
    # Read in the config file
    with open(config_path) as file:
        _cfg: dict = json.load(file)
    config = Config(**_cfg)

    if pm is None:
        pm = ProgressManager()

    # Set up the logger without a file
    logger = setup_rich_logger(logfile=None)

    # Update the amplicon size if it is provided
    if ampliconsizemax:
        logger.info(
            f"Updating amplicon size max to {ampliconsizemax} and min to {ampliconsizemin}"
        )

        config.amplicon_size_max = ampliconsizemax
        config.amplicon_size_min = ampliconsizemin

    # If more than two pools are given throw error
    if config.n_pools > 2:
        raise UsageError("ERROR: repair is only supported with two pools")

    # Create a mapping of chromname/reference to msa_index
    msa_chrom_to_index: dict[str, int] = {
        msa_data["msa_chromname"]: msa_index
        for msa_index, msa_data in _cfg["msa_data"].items()
    }

    # Read in the bedfile
    bedprimerpairs, headers = read_in_bedprimerpairs(primerbed)
    # Map each primer to an MSA index
    for primerpair in bedprimerpairs:
        msa_index = msa_chrom_to_index.get(str(primerpair.chrom_name), None)
        if msa_index is not None:
            primerpair.msa_index = msa_index
        elif _cfg["bedfile"]:
            # This case can happen when primers are added to the scheme via --bedfile.
            # Set the msa index to -1
            primerpair.msa_index = -1
        else:
            raise UsageError(f"ERROR: {primerpair.chrom_name} not found in MSA data")

    bedprimerpairs.sort(key=lambda x: (x.chrom_name, x.amplicon_number))

    # Extract the stem from the primername
    try:
        prefix, ampliconnumber = primername.split("_")[:2]
        primerstem = f"{ampliconnumber}_{prefix}"
    except ValueError:
        raise UsageError(
            f"ERROR: {primername} cannot be parsed using _ as delim"
        ) from None

    # Find primernumber from bedfile
    wanted_pp = None
    for pp in bedprimerpairs:
        if pp.match_primer_stem(primerstem):
            wanted_pp = pp
    if wanted_pp is None:
        raise UsageError(f"ERROR: {primername} not found in bedfile")
    else:
        print(wanted_pp.__str__())

    # Read in the MSAs from config["msa_data"]
    msa_data = _cfg["msa_data"].get(wanted_pp.msa_index)
    if msa_data is None:
        if wanted_pp.msa_index == -1:
            raise UsageError(f"ERROR: The Primer {primername} was added via --bedfile")
        else:
            raise UsageError(f"ERROR: MSA index {wanted_pp.msa_index} not found")

    msa = MSA(
        name=msa_data["msa_name"],
        path=msapath,
        msa_index=wanted_pp.msa_index,
        mapping=config.mapping.value,
        logger=None,
        progress_manager=pm,
    )
    # Check the hashes match
    with open(msa.path, "rb") as f:
        msa_checksum = hashlib.file_digest(f, "md5").hexdigest()
    if msa_checksum != msa_data["msa_checksum"]:
        raise ValueError("ERROR: MSA checksums do not match.")
    else:
        print("MSA checksums match")

    # Target digest the MSA
    msa_fp_end = msa._ref_to_msa[wanted_pp.fprimer.region()[1]]
    msa_rp_start = msa._ref_to_msa[wanted_pp.fprimer.region()[0]]

    findexes = sorted(
        [
            x
            for x in range(
                msa_fp_end - config.amplicon_size_max,
                msa_fp_end + config.amplicon_size_max,
            )
            if x >= 0 and x < msa.array.shape[1]
        ]
    )
    rindexes = sorted(
        [
            x
            for x in range(
                msa_rp_start - config.amplicon_size_max,
                msa_rp_start + config.amplicon_size_max,
            )
            if x >= 0 and x < msa.array.shape[1]
        ]
    )

    # Targeted digestion leads to a mismatch of the indexes.
    # Digest the MSA into FKmers and RKmers
    msa.digest(config, (findexes, rindexes))  ## Primer are remapped at this point.
    print(f"Found {len(msa.fkmers)} FKmers and {len(msa.rkmers)} RKmers")

    # Generate all primerpairs then interaction check
    msa.generate_primerpairs(
        amplicon_size_max=config.amplicon_size_max,
        amplicon_size_min=config.amplicon_size_min,
        dimerscore=config.dimer_score,
    )

    # Find the primers on either side of the wanted primer
    if wanted_pp.amplicon_number == 0:
        left_pp = None
        cov_start = wanted_pp.rprimer.region()[0]
    else:
        left_pp = [
            x
            for x in bedprimerpairs
            if x.amplicon_number == wanted_pp.amplicon_number - 1
        ][0]
        cov_start = left_pp.rprimer.region()[0]

    if wanted_pp.amplicon_number == max([x.amplicon_number for x in msa.primerpairs]):
        right_pp = None
        cov_end = wanted_pp.fprimer.region()[1]
    else:
        right_pp = [
            x
            for x in bedprimerpairs
            if x.amplicon_number == wanted_pp.amplicon_number + 1
        ][0]
        cov_end = right_pp.fprimer.region()[1]

    # Find primerpairs that span the gap
    spanning_primerpairs = [
        x
        for x in msa.primerpairs
        if x.fprimer.region()[1] < cov_start - config.min_overlap
        and x.rprimer.region()[0] > cov_end + config.min_overlap
    ]

    if len(spanning_primerpairs) > 0:
        logger.info(f"Spanning amplicons found: {len(spanning_primerpairs)}")
    else:
        raise ValueError("ERROR: No spanning primers found")

    # Sort for number of primer pairs
    spanning_primerpairs.sort(key=lambda x: len(x.all_seqs()))

    # Find all primerpairs that in the same pool as the wanted primer
    same_pool_primerpairs = [x for x in bedprimerpairs if x.pool == wanted_pp.pool]
    seqs_in_same_pool = [
        seq for seq in (x.all_seqs() for x in same_pool_primerpairs) for seq in seq
    ]

    # Find all primerpairs that in the same pool and same msa as the wanted primer
    spanning_pool_primerpairs = [
        x
        for x in msa.primerpairs
        if x.pool == wanted_pp.pool and x.msa_index == wanted_pp.msa_index
    ]

    accepted_primerpairs = []
    for pos_primerpair in spanning_primerpairs:
        # Make sure the new primerpair doesn't contain the primers we want to replace
        if (
            pos_primerpair.fprimer == wanted_pp.fprimer
            or pos_primerpair.rprimer == wanted_pp.rprimer
        ):
            continue

        # Check for all clashes
        clash = False
        # If they overlap
        for current_pp in spanning_pool_primerpairs:
            # If clash with any skip
            if range(
                max(pos_primerpair.fprimer.region()[0], current_pp.fprimer.region()[0]),
                min(pos_primerpair.rprimer.region()[1], current_pp.rprimer.region()[1])
                + 1,
            ):
                clash = True
                break
        if clash:
            continue

        # If they dont overlap
        # Check for interactions
        if do_pools_interact_py(
            list(pos_primerpair.all_seqs()), seqs_in_same_pool, config.dimer_score
        ):
            continue

        pos_primerpair.pool = wanted_pp.pool
        pos_primerpair.amplicon_number = wanted_pp.amplicon_number

        accepted_primerpairs.append(pos_primerpair)

    accepted_primerpairs.sort(key=lambda a: a.fprimer.end)

    logger.info(f"Found {len(accepted_primerpairs)} valid replacement amplicons")
    for pp_number, pp in enumerate(accepted_primerpairs, 1):
        print(f"Amplicon {pp_number}: {len(pp.all_seqs())} Primers")
        print(pp.to_bed())


def schemecreate(
    msa: list[pathlib.Path],
    output_dir: pathlib.Path,
    config: Config,
    pm: ProgressManager | None,
    force: bool = False,
    input_bedfile: pathlib.Path | None = None,
    offline_plots: bool = True,
):
    """
    Creates a scheme based on multiple sequence alignments (MSA).

    Args:
        msa (list[pathlib.Path]): A list of paths to MSA files.
        output_dir (pathlib.Path): The directory where the scheme will be saved.
        config (Config): Configuration settings for scheme creation.
        pm (ProgressManager | None): Optional. A progress manager instance for UI feedback.
        force (bool): If True, existing output directories will be overwritten.
        input_bedfile (pathlib.Path | None): Optional. A path to an input BED file for incorporating specific primer pairs.
        offline_plots (bool): If True, plots will be generated for offline use.

    Raises:
        SystemExit: If the output directory already exists and the force flag is not set.
    """
    ARG_MSA = msa
    OUTPUT_DIR = pathlib.Path(output_dir).absolute()  # Keep absolute path

    # Create the Config_dict
    cfg_dict = config.to_dict()

    # See if the output dir already exists.
    if OUTPUT_DIR.is_dir() and not force:
        raise UsageError(f"{OUTPUT_DIR} already exists, please use --force to override")

    # Create the output dir and a work subdir
    pathlib.Path.mkdir(OUTPUT_DIR, exist_ok=True)
    pathlib.Path.mkdir(OUTPUT_DIR / "work", exist_ok=True)

    # Set up the logger
    logger = setup_rich_logger(str(OUTPUT_DIR / "work/file.log"))

    if pm is None:
        pm = ProgressManager()

    # Create the mismatch db
    logger.info(
        "Creating the Mismatch Database",
    )
    mismatch_db = MatchDB(
        OUTPUT_DIR / "work/mismatch",
        [str(x) for x in ARG_MSA] if config.use_matchdb else [],
        config.mismatch_kmersize,
    )
    logger.info(
        f"[green]Created[/green]: "
        f"{OUTPUT_DIR.relative_to(OUTPUT_DIR.parent)}/work/mismatch.db",
    )

    # If the bedfile flag is given add the primers into the scheme
    if input_bedfile is not None:
        bedprimerpairs = read_in_extra_primers(input_bedfile, config, logger)

    # Create a dict full of msa data
    msa_data = {}
    msa_dict: dict[int, MSA] = {}

    msa_has_primerpairs_bool = {msa_index: False for msa_index, _ in enumerate(ARG_MSA)}

    for msa_index, msa_path in enumerate(ARG_MSA):
        msa_data[msa_index] = {}

        # Read in the MSA
        msa_obj = MSA(
            name=msa_path.stem,
            path=msa_path,
            msa_index=msa_index,
            mapping=config.mapping.value,
            logger=logger,
            progress_manager=pm,
        )

        # copy the msa into the output / work dir
        local_msa_path = OUTPUT_DIR / "work" / msa_path.name
        msa_obj.write_msa_to_file(local_msa_path)

        # Create MSA checksum
        with open(local_msa_path, "rb") as f:
            msa_data[msa_index]["msa_checksum"] = hashlib.file_digest(
                f, "md5"
            ).hexdigest()

        logger.info(
            f"Read in MSA: [blue]{msa_obj._chrom_name}[/blue]\t"
            f"seqs:[green]{msa_obj.array.shape[0]}[/green]\t"
            f"cols:[green]{msa_obj.array.shape[1]}[/green]"
        )

        # Add some msa data to the dict
        msa_data[msa_index]["msa_name"] = msa_obj.name
        msa_data[msa_index]["msa_path"] = str(
            "work/" + msa_path.name
        )  # Write local path
        msa_data[msa_index]["msa_chromname"] = msa_obj._chrom_name
        msa_data[msa_index]["msa_uuid"] = msa_obj._uuid

        # Add the msa to the dict
        msa_dict[msa_index] = msa_obj

    # Check for collisions in the MSA._chrom_names names
    if len({msa_obj._chrom_name for msa_obj in msa_dict.values()}) != len(msa_dict):
        logger.critical("Duplicate chrom names found in MSA data. Exiting.")
        raise UsageError("Duplicate chrom names found in MSA data")

    # Read in all MSAs before digestion
    for msa_index, msa_obj in msa_dict.items():
        # Digest the MSA into FKmers and RKmers
        msa_obj.digest(config)
        logger.info(
            f"[blue]{msa_obj._chrom_name}[/blue]: digested to "
            f"[green]{len(msa_obj.fkmers)}[/green] FKmers and "
            f"[green]{len(msa_obj.rkmers)}[/green] RKmers"
        )

        # Add the msa to the scheme
        msa_dict[msa_index] = msa_obj

        if len(msa_obj.fkmers) == 0 or len(msa_obj.rkmers) == 0:
            logger.critical(
                f"No valid FKmers or RKmers found for [blue]{msa_obj._chrom_name}[/blue]"
            )
            continue

        # Generate all primerpairs then interaction check
        msa_obj.generate_primerpairs(
            amplicon_size_max=config.amplicon_size_max,
            amplicon_size_min=config.amplicon_size_min,
            dimerscore=config.dimer_score,
        )
        logger.info(
            f"[blue]{msa_obj._chrom_name}[/blue]: Generated "
            f"[green]{len(msa_obj.primerpairs)}[/green] possible amplicons"
        )

        if len(msa_obj.primerpairs) == 0:
            logger.critical(
                f"No valid primers found for [blue]{msa_obj._chrom_name}[/blue]"
            )
        else:
            msa_has_primerpairs_bool[msa_index] = True

    # Add MSA data into cfg
    cfg_dict["msa_data"] = msa_data

    # If all MSAs have no primers, exit
    if not any(msa_has_primerpairs_bool.values()):
        logger.critical("No valid primers found in any MSA")
        raise DigestionFailNoPrimerPairs("No valid primerpairs found in any MSA")

    # Create the scheme object early
    scheme = Scheme(config=config, matchDB=mismatch_db, msa_dict=msa_dict)

    msa_chrom_to_index: dict[str, int] = {
        msa._chrom_name: msa_index for msa_index, msa in msa_dict.items()
    }
    # Add the bedprimerpairs into the scheme
    if input_bedfile is not None and bedprimerpairs:  # type: ignore
        # if input_bedfile != None then bedprimerpairs is assigned
        for bedpp in bedprimerpairs:
            # Map the primerpair to the msa via chromname
            bedpp.msa_index = msa_chrom_to_index.get(bedpp.chrom_name, -1)  # type: ignore
            scheme.add_primer_pair_to_pool(bedpp, bedpp.pool, bedpp.msa_index)
            logger.debug(
                f"Added {bedpp.amplicon_prefix} from [blue]{input_bedfile.name}[/blue]",
            )

    # Start the Scheme generation
    for msa_index, msa_obj in msa_dict.items():
        # Set up the pm for the MSA
        scheme_pt = pm.create_sub_progress(
            iter=None, chrom=msa_obj.name, process="Creating Scheme", leave=False
        )
        scheme_pt.manual_update(n=0, total=msa_obj.array.shape[1])

        while True:
            # Provide the coverage to the progress tracker
            coverage = scheme.get_coverage_percent(msa_index)
            scheme_pt.manual_update(count=coverage if coverage is not None else 0)
            # Update the progress tracker to the current state of the walk
            if scheme._last_pp_added:
                scheme_pt.manual_update(n=scheme._last_pp_added[-1].rprimer.region()[1])

            match scheme.try_ol_primerpairs(msa_obj.primerpairs, msa_index):
                case SchemeReturn.ADDED_OL_PRIMERPAIR:
                    last_pp_added = scheme._last_pp_added[-1]
                    logger.info(
                        "Added [green]overlapping[/green] amplicon for "
                        f"[blue]{msa_obj._chrom_name}[/blue]: {last_pp_added.fprimer.region()[0]}\t"
                        f"{last_pp_added.rprimer.region()[1]}\t{last_pp_added.pool + 1}"
                    )
                    continue
                case SchemeReturn.ADDED_FIRST_PRIMERPAIR:
                    last_pp_added = scheme._last_pp_added[-1]
                    logger.info(
                        "Added [green]first[/green] amplicon for "
                        f"[blue]{msa_obj._chrom_name}[/blue]: {last_pp_added.fprimer.region()[0]}\t"
                        f"{last_pp_added.rprimer.region()[1]}\t{last_pp_added.pool + 1}",
                    )
                    continue
                case SchemeReturn.NO_OL_PRIMERPAIR:
                    pass  # Do nothing move on to next step
                case SchemeReturn.NO_FIRST_PRIMERPAIR:
                    logger.warning(
                        f"No valid primerpairs found for [blue]{msa_obj._chrom_name}[/blue]",
                    )
                    break

            # Try to backtrack
            if config.backtrack:
                logger.info("Backtracking...")
                match scheme.try_backtrack(msa_obj.primerpairs, msa_index):
                    case SchemeReturn.ADDED_BACKTRACKED:
                        last_pp_added = scheme._last_pp_added[-1]
                        logger.info(
                            f"Backtracking allowed [green]overlapping[/green] amplicon for [blue]{msa_obj._chrom_name}[/blue]: "
                            f"{last_pp_added.fprimer.region()[0]}\t{last_pp_added.rprimer.region()[1]}\t{last_pp_added.pool + 1}",
                        )
                    case SchemeReturn.NO_BACKTRACK:
                        logger.info(
                            f"Could not backtrack for [blue]{msa_obj._chrom_name}[/blue]",
                        )
                        pass  # Do nothing move on to next step

            # Try and add a walking primer
            match scheme.try_walk_primerpair(msa_obj.primerpairs, msa_index):
                case SchemeReturn.ADDED_WALK_PRIMERPAIR:
                    last_pp_added = scheme._last_pp_added[-1]
                    logger.info(
                        f"Added [yellow]walking[/yellow] amplicon for [blue]{msa_obj._chrom_name}[/blue]: "
                        f"{last_pp_added.fprimer.region()[0]}\t{last_pp_added.rprimer.region()[1]}\t{last_pp_added.pool + 1}",
                    )
                case _:
                    break

        if config.circular:
            match scheme.try_circular(msa):
                case SchemeReturn.ADDED_CIRCULAR:
                    last_pp_added = scheme._last_pp_added[-1]
                    logger.info(
                        f"Added [green]circular[/green] amplicon for [blue]{msa_obj._chrom_name}[/blue]: "
                        f"{last_pp_added.fprimer.region()[0]}\t{last_pp_added.rprimer.region()[1]}\t{last_pp_added.pool + 1}",
                    )
                case SchemeReturn.NO_CIRCULAR:
                    logger.info(
                        f"No [red]circular[/red] amplicon for [blue]{msa_obj._chrom_name}[/blue]"
                    )
        # Close the progress tracker
        scheme_pt.manual_update(n=scheme_pt.total)
        scheme_pt.close()

    # Create the progress tracker for the final steps
    logger.info("Writing output files")
    upload_steps = 8
    upload_pt = pm.create_sub_progress(
        iter=None,
        chrom="file creation",
        process="Creating output files",
        leave=False,
        bar_format="{l_bar}{bar}",
        unit="%",
    )
    upload_pt.manual_update(n=0, total=upload_steps)

    # Write primer bed file
    with open(OUTPUT_DIR / "primer.bed", "w") as outfile:
        primer_bed_str = scheme.to_bed()
        outfile.write(primer_bed_str)
    upload_pt.manual_update(n=1, update=True)

    # Write amplicon bed file
    with open(OUTPUT_DIR / "amplicon.bed", "w") as outfile:
        amp_bed_str = scheme.to_amplicons(trim_primers=False)
        outfile.write(amp_bed_str)
    with open(OUTPUT_DIR / "primertrim.amplicon.bed", "w") as outfile:
        outfile.write(scheme.to_amplicons(trim_primers=True))
    upload_pt.manual_update(n=2, update=True)

    # Write all the consensus sequences to a single file
    with open(OUTPUT_DIR / "reference.fasta", "w") as reference_outfile:
        reference_records = []
        for msa_obj in msa_dict.values():
            if config.mapping == MappingType.FIRST:
                seq_str = generate_reference(msa_obj.array)
            elif config.mapping == MappingType.CONSENSUS:
                seq_str = generate_consensus(msa_obj.array)
            else:
                raise ValueError("Mapping must be 'first' or 'consensus'")

            reference_records.append(
                SeqRecord.SeqRecord(
                    seq=Seq.Seq(seq_str),
                    id=msa_obj._chrom_name,
                )
            )

        SeqIO.write(reference_records, reference_outfile, "fasta")
    upload_pt.manual_update(n=3, update=True)

    # Create all hashes
    ## Generate the bedfile hash, and add it into the config
    primer_md5 = hashlib.md5("\n".join(primer_bed_str).encode()).hexdigest()
    cfg_dict["primer.bed.md5"] = primer_md5

    ## Generate the amplicon hash, and add it into the config
    amp_md5 = hashlib.md5(amp_bed_str.encode()).hexdigest()
    cfg_dict["amplicon.bed.md5"] = amp_md5

    ## Read in the reference file and generate the hash
    with open(OUTPUT_DIR / "reference.fasta") as reference_outfile:
        ref_md5 = hashlib.md5(reference_outfile.read().encode()).hexdigest()
    cfg_dict["reference.fasta.md5"] = ref_md5

    # Write the config dict to file
    with open(OUTPUT_DIR / "config.json", "w") as outfile:
        outfile.write(json.dumps(cfg_dict, sort_keys=True))

    upload_pt.manual_update(n=4, update=True)

    # Create qc coverage data
    qc_data = {}
    for msa_index, msa_obj in msa_dict.items():
        msa_data = {"coverage": scheme.get_coverage_percent(msa_index)}
        msa_data["n_amplicons"] = len(
            [x for x in scheme._last_pp_added if x.msa_index == msa_index]
        )
        msa_data["n_gaps"] = len(scheme.get_coverage_gaps(msa_index))
        qc_data[msa_obj.name] = msa_data

    with open(OUTPUT_DIR / "work" / "qc.json", "w") as outfile:
        outfile.write(json.dumps(qc_data, sort_keys=True))

    ## DO THIS LAST AS THIS CAN TAKE A LONG TIME
    # Writing plot data
    plot_data = generate_all_plotdata(
        list(msa_dict.values()),
        OUTPUT_DIR / "work",
        last_pp_added=scheme._last_pp_added,
    )
    upload_pt.manual_update(n=5, update=True)

    # Write the plot
    with open(OUTPUT_DIR / "plot.html", "w") as outfile:
        outfile.write(
            generate_all_plots_html(plot_data, OUTPUT_DIR, offline_plots=offline_plots)
        )
    upload_pt.manual_update(n=6, update=True)
    with open(OUTPUT_DIR / "primer.html", "w") as outfile:
        for i, msa_obj in enumerate(msa_dict.values()):
            outfile.write(
                primer_mismatch_heatmap(
                    array=msa_obj.array,
                    seqdict=msa_obj._seq_dict,
                    bedfile=OUTPUT_DIR / "primer.bed",
                    offline_plots=True if offline_plots and i == 0 else False,
                )
            )
    upload_pt.manual_update(n=7, update=True)

    # Close the progress tracker
    upload_pt.manual_update(n=upload_steps, update=True)
    sleep(0.5)  # Sleep to allow the progress bar to finish
    upload_pt.close()

    logger.info("Completed Successfully")



# Core imports
import pathlib

from primaldimer_py import (
    calc_at_offset_py,  # type: ignore
)

from primalscheme3.core.bedfiles import read_in_bedlines

MATCHES: dict[tuple, bool] = {
    ("A", "T"): True,
    ("T", "A"): True,
    ("G", "C"): True,
    ("C", "G"): True,
}

SIMPLE_DNA = {"A", "C", "G", "T"}


def create_cigar(seq1: str, seq2: str) -> str:
    cigar = []
    for seq1base, seq2base in zip(seq1, seq2):
        # Guard against non DNA bases
        if seq1base not in SIMPLE_DNA or seq2base not in SIMPLE_DNA:
            cigar.append(" ")
            continue
        # Check if the bases match
        if MATCHES.get((seq1base, seq2base), False):
            cigar.append("|")
        else:
            cigar.append(".")
    return "".join(cigar)


def create_str(seq1: str, seq2: str, offset: int, score: float) -> str:
    """
    Returns a string representing the interaction between two sequences.
    Example:
        offset = -1
        seq1:   AGCATCATGCTAGCT
                ..|.|..|.|.|.|..
        seq2:    AGCATCATGCTAGCT

    :param seq1: The first sequence  in 5'-3' orientation
    :param seq2: The second sequence in 5'-3' orientation
    :param offset: The offset of the sequences
    :return: A string representing the interaction
    """

    seq2 = seq2[::-1]  # Reverse the sequence

    # Add some direction annotation
    seq1 = f"5'-{seq1}-3' >"
    seq2 = f"3'-{seq2}-5'"

    # Put the sequences in the correct offset
    if offset < 0:
        seq2 = " " * abs(offset) + seq2
    elif offset > 0:
        seq1 = " " * offset + seq1

    # Pad the sequences so they are the same length
    max_length = max(len(seq1), len(seq2))
    seq1 = seq1.ljust(max_length)
    seq2 = seq2.ljust(max_length)
    # Create the cigar string
    cigar = create_cigar(seq1, seq2)

    return f"score: {round(score,2)}\n{seq1}\n{cigar}\n{seq2}\n"


def interaction(seq1: str, seq2: str, threshold: float) -> list[str]:
    """
    Find interactions between two sequences based on a given threshold.

    :param seq1 (str): The first sequence.
    :param seq2 (str): The second sequence.
    :param threshold (float): The threshold for the interaction score.
    :return  list[str]: A list of interactions between the two sequences.
    """
    interactions = []

    for offset in range(-(len(seq1) - 2), len(seq2) - len(seq1)):
        score = calc_at_offset_py(seq1, seq2, offset)
        if score <= threshold:
            interactions.append(create_str(seq1, seq2, offset, score))

    return interactions


def visualise_interactions(bedpath: pathlib.Path, threshold: float) -> None:
    """
    Calculate the interaction score between two sequences.
    If the score is less than the threshold
    :param seq1: Bedfile path
    :param threshold: The threshold for the interaction score
    :return: None
    """

    # Read in the bedfile
    bedlines, header = read_in_bedlines(bedpath)

    # Split the bedfile into pools
    pools = [[] for _ in {bedline.pool for bedline in bedlines}]
    for bedline in bedlines:
        pools[bedline.pool].append(bedline)

    tested = 0
    interactions = 0

    # Loop over the pools
    for pool in pools:
        # Loop over the bedlines in the pool
        for bedline1 in pool:
            for bedline2 in pool:
                tested += 1
                for line in interaction(
                    bedline1.sequence, bedline2.sequence, threshold
                ):
                    interactions += 1
                    print(bedline1.primername, bedline2.primername)
                    print(line)

    print(
        f"Tested {tested} possible combinations and found {interactions} interactions"
    )



from collections import Counter

import numpy as np

# Module imports
from primalscheme3.core.seq_functions import extend_ambiguous_base


def create_mapping(
    msa: np.ndarray, mapping_index: int = 0
) -> tuple[np.ndarray, np.ndarray]:
    """
    This returns a tuple of two items:
        mapping_array: list[int | None]
        truncated_msa: np.ndarray
    mapping_array: Each position in the list corresponds to the same index in the MSA, The value in the list is the position in the reference genome
    """
    # As NP is modified in place, returning is not necessary but is done for clarity
    # Create the empty mapping array
    mapping_list = [None] * msa.shape[1]
    mapping_array = np.array(mapping_list)
    # Select the reference genome
    reference_genome = msa[mapping_index]
    # Iterate over the msa genome
    current_ref_index = 0
    for col_index in range(msa.shape[1]):
        # If the base is not a gap, assign the mapping
        if reference_genome[col_index] not in {"", "-"}:
            mapping_array[col_index] = current_ref_index
            # increase reference index
            current_ref_index += 1
    return (mapping_array, msa)


def generate_consensus(msa: np.ndarray) -> str:
    """
    Generates a consensus sequence from an msa
    """
    consensus = []
    # For each column in the msa
    for col in range(msa.shape[1]):
        # Create the counter
        col_counter = Counter()
        # For each row in the msa
        for row in range(msa.shape[0]):
            # Update the counter with the de-ambiguous bases
            col_counter.update(extend_ambiguous_base(msa[row, col]))

        # Remove invalid bases if other bases are available
        col_counter.pop("N", None)

        if len(col_counter) == 0:
            consensus.append("N")
        else:
            consensus.append(col_counter.most_common(1)[0][0])
    return "".join(consensus)


def generate_reference(msa: np.ndarray) -> str:
    """
    Generates a reference string from the first row of a multiple sequence alignment (MSA) array.

    Args:
    - msa (np.ndarray): A numpy array representing a multiple sequence alignment.

    Returns:
    - str: A string representing the reference sequence, obtained by joining the characters in the first row of the MSA and removing any gaps represented by hyphens ("-").
    """

    return "".join(msa[0]).replace("-", "")


def ref_index_to_msa(mapping_array: np.ndarray) -> dict[int, int]:
    """
    Convert a reference index to an MSA index
    """
    return {x: i for i, x in enumerate(list(mapping_array)) if x is not None}


def check_for_end_on_gap(ref_index_to_msa: dict[int, int], ref_index) -> bool:
    """
    Check if a slice of a mapping array ends on a gap
    Returns True if the slice ends on a gap, False otherwise

    # Example

           5' AGAGTGTGGGGGTAGTGTTACG          > MPXV_142_LEFT_1 170931:170953
    TTTTTTTTATAGAGTGTGGGGGTAGTGTTACG-------GAT >MT903345
    TTTTTTTTATAGAGTGT-GGGGTAGTGTTACGGATATCTGAT >KJ642613.1
                                           ^ 173598: MSA
                                           ^ 170953: ref

    In this example, the slice of the primer ends on a gap. So slicing the array with
    - array[:, ref_to_msa[170931]:ref_to_msa[170953]] will return "GGGGTAGTGTTACG-------" as the non exclusive end captures the gap
    - fix_end_on_gap() will return in indexes to slice the array without the gap ie "TGTGGGGGTAGTGTTACG"

    """
    exclusive_msa_end = ref_index_to_msa[ref_index]
    inclusive_msa_end = ref_index_to_msa[ref_index - 1]
    return exclusive_msa_end - inclusive_msa_end != 1


def fix_end_on_gap(ref_index_to_msa: dict[int, int], ref_index) -> int:
    """
    Returns the MSA index of the non-inclusive end of a slice with the gap removed
    """
    return ref_index_to_msa[ref_index - 1] + 1



#!/usr/bin/python3
import argparse
import pathlib
from typing import Optional

import typer
from typing_extensions import Annotated

# Module imports
from primalscheme3.__init__ import __version__
from primalscheme3.core.config import Config, MappingType
from primalscheme3.core.msa import parse_msa
from primalscheme3.core.primer_visual import bedfile_plot_html, primer_mismatch_heatmap
from primalscheme3.core.progress_tracker import ProgressManager
from primalscheme3.interaction.interaction import (
    visualise_interactions,
)
from primalscheme3.panel.panel_main import PanelRunModes, panelcreate
from primalscheme3.repair.repair import repair

# Import main functions
from primalscheme3.scheme.scheme_main import schemecreate, schemereplace

## Commands are in the format of
# {pclass}-{mode}
# pclass = panel or scheme

# Example to create a scheme
# scheme-create

# To repair a scheme
# scheme-repair

# To create a panel
# panel-create


def check_path_is_file(value: str | pathlib.Path) -> pathlib.Path:
    if isinstance(value, str):
        value = pathlib.Path(value)
    if not value.is_file():
        raise argparse.ArgumentTypeError(f"No file found at: '{str(value.absolute())}'")
    return value


# Create the main app
app = typer.Typer(name="primalscheme3", no_args_is_help=True)


def check_output_dir(output: pathlib.Path, force: bool):
    if output.exists() and not force:
        raise typer.BadParameter(
            f"--output '{output}' directory already exists. Use --force to overwrite"
        )


def typer_callback_version(value: bool):
    if value:
        version = typer.style(__version__, fg=typer.colors.GREEN, bold=True)
        typer.echo("PrimalScheme3 version: " + version)
        raise typer.Exit()


@app.callback()
def primalscheme3(
    value: Annotated[bool, typer.Option] = typer.Option(
        False, "--version", callback=typer_callback_version
    ),
):
    pass


@app.command(no_args_is_help=True)
def scheme_create(
    msa: Annotated[
        list[pathlib.Path],
        typer.Option(
            help="The MSA to design against. To use multiple MSAs, use multiple --msa flags. (--msa 1.fasta --msa 2.fasta)",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    output: Annotated[
        pathlib.Path,
        typer.Option(
            help="The output directory",
            resolve_path=True,
        ),
    ],
    amplicon_size: Annotated[
        int,
        typer.Option(
            help="The size of an amplicon. Min / max size are ± 10 percent [100<=x<=2000]",
        ),
    ] = Config.amplicon_size,
    bedfile: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="An existing bedfile to add primers to",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ] = None,
    min_overlap: Annotated[
        int,
        typer.Option(help="min amount of overlap between primers", min=0),
    ] = Config.min_overlap,
    n_pools: Annotated[
        int, typer.Option(help="Number of pools to use", min=1)
    ] = Config.n_pools,
    dimer_score: Annotated[
        float,
        typer.Option(
            help="Threshold for dimer interaction",
        ),
    ] = Config.dimer_score,
    min_base_freq: Annotated[
        float,
        typer.Option(help="Min freq to be included,[0<=x<=1]", min=0.0, max=1.0),
    ] = Config.min_base_freq,
    mapping: Annotated[
        MappingType,
        typer.Option(
            help="How should the primers in the bedfile be mapped",
        ),
    ] = Config.mapping.value,  # type: ignore
    circular: Annotated[
        bool, typer.Option(help="Should a circular amplicon be added")
    ] = Config.circular,
    backtrack: Annotated[
        bool, typer.Option(help="Should the algorithm backtrack")
    ] = Config.backtrack,
    ignore_n: Annotated[
        bool,
        typer.Option(help="Should N in the input genomes be ignored"),
    ] = Config.ignore_n,
    force: Annotated[
        bool, typer.Option(help="Override the output directory")
    ] = Config.force,
    input_bedfile: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Path to a primer.bedfile containing the pre-calculated primers"
        ),
    ] = Config.input_bedfile,
    high_gc: Annotated[bool, typer.Option(help="Use high GC primers")] = Config.high_gc,
    offline_plots: Annotated[
        bool,
        typer.Option(
            help="Includes 3Mb of dependencies into the plots, so they can be viewed offline"
        ),
    ] = True,
    use_matchdb: Annotated[
        bool,
        typer.Option(
            help="Create and use a mispriming database",
        ),
    ] = Config.use_matchdb,
):
    """
    Creates a tiling overlap scheme for each MSA file
    """
    # Update the config with CLI params
    config = Config(**locals())

    # Check the output directory
    check_output_dir(output, force)

    # Set up the progress manager
    pm = ProgressManager()
    schemecreate(
        msa=msa,
        output_dir=output,
        config=config,
        pm=pm,
        force=force,
        input_bedfile=input_bedfile,
        offline_plots=offline_plots,
    )


@app.command(no_args_is_help=True)
def scheme_replace(
    primername: Annotated[
        str, typer.Argument(help="The name of the primer to replace")
    ],
    primerbed: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The bedfile containing the primer to replace",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    msa: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The msa used to create the original primer scheme",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    amplicon_size: Annotated[
        int,
        typer.Option(
            help="The size of an amplicon. Use single value for ± 10 percent [100<=x<=2000]",
        ),
    ],
    config: Annotated[
        pathlib.Path,
        typer.Option(
            help="The config.json used to create the original primer scheme",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
):
    """
    Replaces a primerpair in a bedfile
    """
    ampliconsizemin = int(amplicon_size * 0.9)
    ampliconsizemax = int(amplicon_size * 1.1)

    # Set up the progress manager
    pm = ProgressManager()

    schemereplace(
        config_path=config,
        primername=primername,
        ampliconsizemax=ampliconsizemax,
        ampliconsizemin=ampliconsizemin,
        primerbed=primerbed,
        msapath=msa,
        pm=pm,
    )


@app.command(no_args_is_help=True)
def panel_create(
    msa: Annotated[
        list[pathlib.Path],
        typer.Option(
            help="Paths to the MSA files", exists=True, readable=True, resolve_path=True
        ),
    ],
    output: Annotated[
        pathlib.Path,
        typer.Option(
            help="The output directory",
            resolve_path=True,
        ),
    ],
    region_bedfile: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Path to the bedfile containing the wanted regions",
            readable=True,
            dir_okay=False,
            file_okay=True,
            exists=True,
        ),
    ] = None,
    input_bedfile: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Path to a primer.bedfile containing the pre-calculated primers",
            readable=True,
            dir_okay=False,
            file_okay=True,
            exists=True,
        ),
    ] = None,
    mode: Annotated[
        PanelRunModes,
        typer.Option(
            help="Select what run mode",
        ),
    ] = PanelRunModes.REGION_ONLY.value,  # type: ignore
    amplicon_size: Annotated[
        int, typer.Option(help="The size of an amplicon")
    ] = Config.amplicon_size,
    n_pools: Annotated[
        int, typer.Option(help="Number of pools to use", min=1)
    ] = Config.n_pools,
    dimer_score: Annotated[
        float, typer.Option(help="Threshold for dimer interaction")
    ] = Config.dimer_score,
    min_base_freq: Annotated[
        float,
        typer.Option(help="Min freq to be included,[0<=x<=1]", min=0.0, max=1.0),
    ] = Config.min_base_freq,
    mapping: Annotated[
        MappingType,
        typer.Option(
            help="How should the primers in the bedfile be mapped",
        ),
    ] = Config.mapping.value,  # type: ignore
    max_amplicons: Annotated[
        Optional[int], typer.Option(help="Max number of amplicons to create", min=1)
    ] = None,
    max_amplicons_msa: Annotated[
        Optional[int], typer.Option(help="Max number of amplicons for each MSA", min=1)
    ] = None,
    max_amplicons_region_group: Annotated[
        Optional[int],
        typer.Option(help="Max number of amplicons for each region", min=1),
    ] = None,
    force: Annotated[bool, typer.Option(help="Override the output directory")] = False,
    high_gc: Annotated[bool, typer.Option(help="Use high GC primers")] = Config.high_gc,
    offline_plots: Annotated[
        bool,
        typer.Option(
            help="Includes 3Mb of dependencies into the plots, so they can be viewed offline"
        ),
    ] = True,
    use_matchdb: Annotated[
        bool,
        typer.Option(
            help="Create and use a mispriming database",
        ),
    ] = Config.use_matchdb,
):
    """
    Creates a primer panel
    """
    # Update the config with CLI params
    config = Config(**locals())

    # Check the output directory
    check_output_dir(output, force)

    # Set up the progress manager
    pm = ProgressManager()

    panelcreate(
        msa=msa,
        output_dir=output,
        region_bedfile=region_bedfile,
        input_bedfile=input_bedfile,
        mode=mode,
        config=config,
        pm=pm,
        max_amplicons=max_amplicons,
        max_amplicons_msa=max_amplicons_msa,
        max_amplicons_region_group=max_amplicons_region_group,
        force=force,
        offline_plots=offline_plots,
    )


@app.command(no_args_is_help=True)
def interactions(
    bedfile: Annotated[
        pathlib.Path,
        typer.Argument(
            help="Path to the bedfile",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    threshold: Annotated[
        float,
        typer.Option(
            help="Only show interactions more severe (Lower score) than this value",
        ),
    ] = -26.0,
):
    """
    Shows all the primer-primer interactions within a bedfile
    """
    visualise_interactions(bedfile, threshold)


@app.command(no_args_is_help=True)
def repair_mode(
    bedfile: Annotated[
        pathlib.Path,
        typer.Option(
            help="Path to the bedfile",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    msa: Annotated[
        pathlib.Path,
        typer.Option(
            help="An MSA, with the reference.fasta, aligned to any new genomes with mutations",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    config: Annotated[
        pathlib.Path,
        typer.Option(
            help="Path to the config.json",
            exists=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    output: Annotated[
        pathlib.Path, typer.Option(help="The output directory", dir_okay=True)
    ],
    force: Annotated[bool, typer.Option(help="Override the output directory")] = False,
):
    """
    Repairs a primer scheme via adding more primers to account for new mutations
    """
    # Set up the progress manager
    pm = ProgressManager()

    repair(
        config_path=config,
        bedfile_path=bedfile,
        force=force,
        pm=pm,
        output_dir=output,
        msa_path=msa,
    )


@app.command(no_args_is_help=True)
def visualise_primer_mismatches(
    msa: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The MSA used to design the scheme",
            readable=True,
            exists=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    bedfile: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The bedfile containing the primers",
            readable=True,
            exists=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    output: Annotated[
        pathlib.Path,
        typer.Option(help="Output location of the plot", dir_okay=False, writable=True),
    ] = pathlib.Path("primer.html"),
    include_seqs: Annotated[
        bool,
        typer.Option(help="Reduces plot filesize, by excluding primer sequences"),
    ] = True,
    offline_plots: Annotated[
        bool,
        typer.Option(
            help="Includes 3Mb of dependencies into the plots, so they can be viewed offline"
        ),
    ] = True,
):
    """
    Visualise mismatches between primers and the input genomes
    """

    array, seqdict = parse_msa(msa)

    with open(output, "w") as outfile:
        outfile.write(
            primer_mismatch_heatmap(
                array=array,
                seqdict=seqdict,
                bedfile=bedfile,
                offline_plots=offline_plots,
                include_seqs=include_seqs,
            )
        )


@app.command(no_args_is_help=True)
def visualise_bedfile(
    bedfile: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The bedfile containing the primers",
            readable=True,
            exists=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    ref_id: Annotated[str, typer.Option(help="The reference genome ID")],
    ref_path: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The bedfile containing the primers",
            readable=True,
            exists=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    output: Annotated[
        pathlib.Path,
        typer.Option(help="Output location of the plot", dir_okay=False, writable=True),
    ] = pathlib.Path("bedfile.html"),
):
    """
    Visualise the bedfile
    """
    from Bio import SeqIO

    ref_file = SeqIO.index(ref_path, "fasta")
    ref_genome = ref_file.get(ref_id)
    if ref_genome is None:
        raise typer.BadParameter(
            f"Reference genome ID '{ref_id}' not found in '{ref_path}'"
        )

    with open(output, "w") as outfile:
        outfile.write(
            bedfile_plot_html(bedfile=bedfile, ref_name=ref_id, ref_seq=ref_genome)
        )


if __name__ == "__main__":
    app()



from collections import Counter
from itertools import product
from math import log2
from typing import Iterable

import numpy as np

# Module Imports
from primalscheme3.core.config import (
    ALL_BASES,
    ALL_DNA,
    AMB_BASES,
    AMBIGUOUS_DNA_COMPLEMENT,
    SIMPLE_BASES,
)


def reverse_complement(kmer_seq: str) -> str:
    rev_seq = kmer_seq[::-1]
    return "".join(AMBIGUOUS_DNA_COMPLEMENT[base.upper()] for base in rev_seq)


def get_most_common_base(array: np.ndarray, col_index: int) -> str:
    values, counts = np.unique(array[:, col_index], return_counts=True)
    counter = dict(zip(values, counts))
    counter.pop("", None)
    return max(counter, key=counter.get)  # type: ignore


def remove_end_insertion(msa_array: np.ndarray) -> np.ndarray:
    """
    Removes leading and trailing "-" from an msa
    """
    tmp_array = msa_array
    ncols = tmp_array.shape[1]
    for row_index in range(0, tmp_array.shape[0]):
        # Solves the 5' end
        for col_index in range(0, ncols):
            if tmp_array[row_index, col_index] == "-":
                tmp_array[row_index, col_index] = ""
            else:
                break
        for rev_col_index in range(ncols - 1, 0, -1):
            if tmp_array[row_index, rev_col_index] == "-":
                tmp_array[row_index, rev_col_index] = ""
            else:
                break
    return tmp_array


def expand_ambs(seqs: Iterable[str]) -> set[str] | None:
    """
    Takes a list / set of strings and returns a set with all ambs expanded
    Return None on invalid bases (Including N)
    """
    returned_seq = set()

    for seq in seqs:
        bases = {*seq}

        # if invalid bases are in sequence return None
        if not bases.issubset(ALL_BASES):
            return None

        # If there is any amb_bases in the seq
        if bases & AMB_BASES:
            expanded_seqs = set(map("".join, product(*map(ALL_DNA.get, seq))))  # type: ignore
            for exp_seq in expanded_seqs:
                returned_seq.add(exp_seq)
        else:
            returned_seq.add(seq)
    return returned_seq


def extend_ambiguous_base(base: str) -> list[str]:
    """Return list of all possible sequences given an ambiguous DNA input"""
    return [*ALL_DNA.get(base, "N")]


def calc_entropy(probs: list[float]) -> float:
    return -sum([p * log2(p) for p in probs if p])


def calc_probs(bases: list[str]) -> list[float]:
    """Calculate the probability/proportion of each base in each column"""
    all_bases = [
        y for sublist in (extend_ambiguous_base(x) for x in bases) for y in sublist
    ]
    counter = Counter(all_bases)
    num_invalids = counter.pop("N", 0)

    return [v / (len(all_bases) - num_invalids) for _, v in counter.items()]


def entropy_score_array(msa: np.ndarray) -> list[float]:
    """
    Creates an list with the entropy at each index
    """
    score_array: list[float] = [0] * msa.shape[1]
    # Iterate over colums
    for col in range(msa.shape[1]):
        value, counts = np.unique(msa[:, col], return_counts=True)
        count_dict = dict(zip(value, counts))

        # Remove non DNA bases
        count_dict = {k: v for k, v in count_dict.items() if k in ALL_BASES}

        parsed_counts = {base: 0 for base in "ACGT"}
        # Expand ambiguous bases
        for base in count_dict:
            if base in AMB_BASES:
                amb_count = count_dict.get(base, 0)  # Remove the ambiguous base
                for expanded_base in extend_ambiguous_base(base):
                    if expanded_base != "N":
                        parsed_counts[expanded_base] += amb_count
            else:
                parsed_counts[base] += count_dict.get(base, 0)
        # Remove Invalid bases
        parsed_counts = {k: v for k, v in count_dict.items() if k in SIMPLE_BASES}

        # Calculate the proportions (probabilities) of each base
        proportions = [v / sum(parsed_counts.values()) for v in parsed_counts.values()]

        score_array[col] = calc_entropy(proportions)
    return score_array



# Core imports
import hashlib
import json

# General import
import pathlib
import shutil
import sys

from Bio import Seq, SeqIO, SeqRecord
from click import UsageError

# version import
from primalscheme3.core.bedfiles import read_in_extra_primers
from primalscheme3.core.config import Config, MappingType
from primalscheme3.core.create_report_data import generate_all_plotdata
from primalscheme3.core.create_reports import generate_all_plots_html
from primalscheme3.core.logger import setup_rich_logger
from primalscheme3.core.mapping import generate_consensus, generate_reference
from primalscheme3.core.mismatches import MatchDB
from primalscheme3.core.primer_visual import primer_mismatch_heatmap
from primalscheme3.core.progress_tracker import ProgressManager

# Module imports
from primalscheme3.panel.panel_classes import (
    Panel,
    PanelMSA,
    PanelReturn,
    PanelRunModes,
    Region,
    RegionParser,
)


def mean_gc_diff(seqs: list[str] | set[str], target_gc=0.5) -> float:
    gc_diff = []
    for seq in seqs:
        gc_diff.append(abs(target_gc - ((seq.count("G") + seq.count("C")) / len(seq))))
    return sum(gc_diff) / len(seqs)


def read_region_bedfile(path) -> list[list[str]]:
    """
    Bedfiles need to be in the format:
    chrom start end name score
    """
    ## If bedfile given, parse it:
    bed_lines = []
    with open(path) as bedfile:
        for line in bedfile.readlines():
            line = line.strip()
            if not line or line.startswith("#"):  # Skip empty lines and header lines
                continue
            bed_lines.append(line.split("\t"))
    return bed_lines


def panelcreate(
    msa: list[pathlib.Path],
    output_dir: pathlib.Path,
    config: Config,
    pm: ProgressManager | None,
    force: bool = False,
    input_bedfile: pathlib.Path | None = None,
    region_bedfile: pathlib.Path | None = None,
    mode: PanelRunModes = PanelRunModes.ENTROPY,
    max_amplicons: int | None = None,
    max_amplicons_msa: int | None = None,
    max_amplicons_region_group: int | None = None,
    offline_plots: bool = True,
):
    ARG_MSA = msa
    OUTPUT_DIR = pathlib.Path(output_dir).absolute()

    # Config Dicts
    config_dict = config.to_dict()
    config_dict["max_amplicons"] = max_amplicons
    config_dict["mode"] = mode.value
    config_dict["max_amplicons_msa"] = max_amplicons_msa
    config_dict["max_amplicons_region_group"] = max_amplicons_region_group

    # Enforce mapping
    if config.mapping != MappingType.FIRST:
        raise UsageError("mapping must be 'first'")

    # Enforce region only has a region bedfile
    if mode == PanelRunModes.REGION_ONLY and region_bedfile is None:
        raise UsageError(
            "region-only mode requires a region bedfile. Provide with --region-bedfile"
        )

    # See if the output dir already exists
    if OUTPUT_DIR.is_dir() and not force:
        raise UsageError(f"{OUTPUT_DIR} already exists, please use --force to override")

    # Create the output dir and a work subdir
    pathlib.Path.mkdir(OUTPUT_DIR, exist_ok=True)
    pathlib.Path.mkdir(OUTPUT_DIR / "work", exist_ok=True)

    ## Set up the logger
    logger = setup_rich_logger(str(OUTPUT_DIR / "work" / "file.log"))

    # Check
    if mode != PanelRunModes.REGION_ONLY and max_amplicons_region_group is not None:
        logger.warning(
            "max-amplicons-group is set but mode is not region-only. Ignoring"
        )
        max_amplicons_region_group = None

    ## Set up the progress manager
    if pm is None:
        pm = ProgressManager()

    # Create the mismatch db
    logger.info(
        "Creating the Mismatch Database",
    )
    mismatch_db = MatchDB(
        OUTPUT_DIR / "work/mismatch",
        [str(x) for x in ARG_MSA] if config.use_matchdb else [],
        config.mismatch_kmersize,
    )
    logger.info(
        f"[green]Created[/green]: "
        f"{OUTPUT_DIR.relative_to(OUTPUT_DIR.parent)}/work/mismatch.db",
    )

    regions_mapping: dict[Region, str | None] | None = None
    # Read in the region_bedfile if given
    if region_bedfile is not None:
        region_bed_lines = read_region_bedfile(region_bedfile)

        regions_mapping = {
            RegionParser.from_list(bed_list): None for bed_list in region_bed_lines
        }
        try:
            shutil.copy(region_bedfile, OUTPUT_DIR / region_bedfile.name)  # type: ignore # Copy the bedfile
        except shutil.SameFileError:
            pass

    # If the bedfile flag is given add the primers into the scheme
    if input_bedfile is not None:
        bedprimerpairs = read_in_extra_primers(input_bedfile, config, logger)

    ## Read in the MSAs
    msa_dict: dict[int, PanelMSA] = {}
    msa_data: dict = {}
    for msa_index, msa_path in enumerate(ARG_MSA):
        msa_data[msa_index] = {}

        # Read in the MSA
        msa_obj = PanelMSA(
            name=msa_path.stem,
            path=msa_path,
            msa_index=msa_index,
            mapping=config.mapping.value,
            logger=logger,
            progress_manager=pm,
        )

        # copy the msa into the output / work dir
        local_msa_path = OUTPUT_DIR / "work" / msa_path.name
        msa_obj.write_msa_to_file(local_msa_path)

        # Create MSA checksum
        with open(local_msa_path, "rb") as f:
            msa_data[msa_index]["msa_checksum"] = hashlib.file_digest(
                f, "md5"
            ).hexdigest()

        logger.info(
            f"Read in MSA: [blue]{local_msa_path.name}[/blue] ({msa_obj._chrom_name})\t"
            f"seqs:[green]{msa_obj.array.shape[0]}[/green]\t"
            f"cols:[green]{msa_obj.array.shape[1]}[/green]"
        )
        # Add the MSA to the dict
        msa_dict[msa_index] = msa_obj

        # Add some msa data to the dict
        msa_data[msa_index]["msa_name"] = msa_obj.name
        msa_data[msa_index]["msa_path"] = str(
            "work/" + msa_path.name
        )  # Write local path
        msa_data[msa_index]["msa_chromname"] = msa_obj._chrom_name
        msa_data[msa_index]["msa_uuid"] = msa_obj._uuid

        # Add the regions
        msa_regions = None
        if regions_mapping is not None:
            msa_regions = []
            for region in regions_mapping.keys():
                if region.chromname == msa_obj._chrom_name:
                    msa_regions.append(region)
                    regions_mapping[region] = msa_obj._chrom_name

            # Print Number mapped
            logger.info(
                f"[blue]{msa_obj._chrom_name}[/blue]: "
                f"[green]{len(msa_regions)}[/green] regions mapped",
            )

        # is mode is all msa_regions is None
        # Creates the score array
        msa_obj.add_regions(msa_regions)
        msa_obj.create_score_array(
            msa_regions, mode=mode
        )  # if mode is not REGION_ONLY, msa_regions is None

    # Print mapping stats
    if regions_mapping is not None:
        mapped_regions = sum(1 for x in regions_mapping.values() if x is not None)
        unmapped_regions = len(regions_mapping) - mapped_regions
        logger.info(
            f"Regions mapped: ([green]{mapped_regions}[/green]). Regions not mapped: ({unmapped_regions})",
        )
        if unmapped_regions > 0:
            logger.warning(
                f"Regions not mapped: {set([x.chromname for x, y in regions_mapping.items() if y is None])}"
            )
        if mapped_regions == 0:
            logger.critical(
                f"No regions mapped. Input genomes use chrom names ({', '.join(x._chrom_name for x in msa_dict.values())}). Regions use chrom names ({', '.join(set(x.chromname for x in regions_mapping.keys()))})"
            )
            sys.exit(1)

        # Check all genomes have regions
        for msa_obj in msa_dict.values():
            if msa_obj._chrom_name not in regions_mapping.values():  # type: ignore
                logger.critical(
                    f"Provided genome ({msa_obj._chrom_name}) has no regions. Please remove genome or add regions"
                )
                sys.exit(1)

    # Start the digestion loop
    for _msa_index, msa_obj in msa_dict.items():
        if msa_obj.regions is not None:
            indexes = set()
            for region in msa_obj.regions:
                indexes.update(
                    range(
                        msa_obj._ref_to_msa[region.start],
                        msa_obj._ref_to_msa[region.stop],
                    )
                )
            # Clean up
            findexes = sorted(
                {
                    fi
                    for fi in (
                        range(i - config.amplicon_size_max, i + 1) for i in indexes
                    )
                    for fi in fi
                    if fi >= 0 and fi < msa_obj.array.shape[1]
                }
            )
            rindexes = sorted(
                {
                    ri
                    for ri in (range(i, i + config.amplicon_size_max) for i in indexes)
                    for ri in ri
                    if ri >= 0 and ri < msa_obj.array.shape[1]
                }
            )
        # Split the logic for the different modes
        logger.info(
            f"Digesting [blue]{msa_obj._chrom_name}[/blue]",
        )
        match mode:
            case PanelRunModes.REGION_ONLY:
                msa_obj.digest(config=config, indexes=(findexes, rindexes))  # type: ignore
            case _:
                msa_obj.digest(config=config, indexes=None)

        # Log the digestion
        logger.info(
            f"[blue]{msa_obj._chrom_name}[/blue]: digested to "
            f"[green]{len(msa_obj.fkmers)}[/green] FKmers and "
            f"[green]{len(msa_obj.rkmers)}[/green] RKmers"
        )
        if len(msa_obj.fkmers) == 0 or len(msa_obj.rkmers) == 0:
            logger.critical(
                f"No valid FKmers or RKmers found for [blue]{msa_obj._chrom_name}[/blue]"
            )
            continue

        # Generate all primerpairs
        msa_obj.generate_primerpairs(
            amplicon_size_max=config.amplicon_size_max,
            amplicon_size_min=config.amplicon_size_max,
            dimerscore=config.dimer_score,
        )
        logger.info(
            f"[blue]{msa_obj._chrom_name}[/blue]: Generated "
            f"[green]{len(msa_obj.primerpairs)}[/green] possible amplicons"
        )
        if len(msa_obj.primerpairs) == 0:
            logger.critical(
                f"No valid primers found for [blue]{msa_obj._chrom_name}[/blue]"
            )

        match mode:
            case PanelRunModes.REGION_ONLY:
                # Filter the primerpairs to only include the ones with scores (cover regions)
                msa_obj.primerpairs = [
                    x for x in msa_obj.primerpairs if msa_obj.get_pp_score(x) > 0
                ]
            case _:
                continue

    # Add all the msa_data to the cfg
    config_dict["msa_data"] = msa_data

    ## Digestion finished, now create the panel

    # Create the regions for the regions_mapping
    chrom_name_regions_mapping: dict[str, dict[str, list[Region]]] = {}
    if regions_mapping is not None:
        for region, chrom_name in regions_mapping.items():
            # Add only mapped regions
            if chrom_name is None:
                continue
            if chrom_name not in chrom_name_regions_mapping:
                chrom_name_regions_mapping[chrom_name] = {}

            if region.name not in chrom_name_regions_mapping[chrom_name]:
                chrom_name_regions_mapping[chrom_name][region.name] = []
            chrom_name_regions_mapping[chrom_name][region.name].append(region)

    # Create a lookup dict for the msa index to name
    msa_index_to_name = {k: v._chrom_name for k, v in msa_dict.items()}

    # Create a dict to store how many amplicons have been added to each msa
    msa_index_to_amplicon_count = {k: 0 for k in msa_data.keys()}

    # Create the panel object
    panel: Panel = Panel(msa_dict, config=config, matchdb=mismatch_db, logger=logger)

    # MSA_INDEX_TO_CHROMNAME =
    msa_chromname_to_index = {
        msa._chrom_name: msa.msa_index for msa in msa_dict.values()
    }

    # Read in the input_bedfile if given
    if input_bedfile is not None:
        for bedpp in bedprimerpairs:  # type: ignore
            bedpp.msa_index = msa_chromname_to_index.get(
                bedpp.chrom_name,  # type: ignore
                -1,
            )
            panel._add_primerpair(
                bedpp,
                pool=bedpp.pool,
                msa_index=bedpp.msa_index,  # type: ignore
            )
            logger.debug(
                f"Added {bedpp.amplicon_prefix} from "
                f"[blue]{input_bedfile.name}[/blue]",
            )

    # Add the first primerpair
    counter = 0

    # Randomise the order of primerpairs, sorting them by fprimer hash
    for msa_obj in msa_dict.values():
        msa_obj.primerpairs = sorted(
            msa_obj.primerpairs, key=lambda x: x.fprimer.__hash__()
        )

    while max_amplicons is None or counter < max_amplicons:
        # Check if the panel is complete
        if all(panel._is_msa_index_finished.values()):
            break

        match panel.add_next_primerpair(max_amplicons_group=max_amplicons_region_group):
            case PanelReturn.ADDED_PRIMERPAIR:
                added_pp = panel._last_pp_added[-1]
                # Update the amplicon count
                msa_index_to_amplicon_count[added_pp.msa_index] += 1
                logger.info(
                    f"Added amplicon ([green]{msa_index_to_amplicon_count.get(added_pp.msa_index)}[/green]) "
                    f"for [blue]{msa_index_to_name.get(added_pp.msa_index)}[/blue]: "
                    f"{added_pp.fprimer.region()[0]}\t{added_pp.rprimer.region()[1]}\t{added_pp.pool + 1}",
                )
                counter += 1

                # If number of amplicons per msa is reached. Set as finished
                if max_amplicons_msa is not None and (
                    msa_index_to_amplicon_count[added_pp.msa_index] >= max_amplicons_msa
                ):
                    panel._is_msa_index_finished[added_pp.msa_index] = True

                chrom_name = msa_index_to_name[added_pp.msa_index]
                continue
            case PanelReturn.NO_PRIMERPAIRS_IN_MSA:
                logger.debug("Skipping MSA")
                continue
            case _:  # pragma: no cover
                logger.error("Unknown return from add_next_primerpair")

    # Log that the panel is finished
    logger.info(
        f"Finished creating the panel. [green]{len(panel._last_pp_added)}[/green] amplicons total",
    )
    # Print the amplicons count for each msa
    for msa_index, amplicon_count in msa_index_to_amplicon_count.items():
        logger.info(
            f"[blue]{msa_index_to_name.get(msa_index)}[/blue]: [green]{amplicon_count}[/green] amplicons",
        )

    region_to_coverage = {}
    # If region bedfile given, check that all regions have been covered
    if region_bedfile is not None:
        for msa_obj in panel._msa_dict.values():
            assert msa_obj.regions is not None
            for region in msa_obj.regions:
                # Skip if region is in a group
                if region.group is not None:
                    continue

                region_coverage = panel._coverage[msa_obj.msa_index][
                    region.start : region.stop
                ]
                region_mean_coverage = region_coverage.mean()
                percent = round(region_mean_coverage * 100, 2)

                log_percent_col = "yellow" if percent < 100 else "green"
                logger.info(
                    f"[blue]{msa_obj._chrom_name}[/blue]:{region.name} "
                    f"{region.start}:{region.stop} "
                    f"[{log_percent_col}]{percent}%[/{log_percent_col}] covered",
                )
                region_to_coverage[region] = region_mean_coverage

    logger.info(
        "Writing outputs...",
    )
    # Write primer bed file
    with open(OUTPUT_DIR / "primer.bed", "w") as outfile:
        primer_bed_str = panel.to_bed()
        outfile.write(primer_bed_str)

    # Write amplicon bed file
    with open(OUTPUT_DIR / "amplicon.bed", "w") as outfile:
        amp_bed_str = panel.to_amplicons(trim_primers=False)
        outfile.write(amp_bed_str)
    with open(OUTPUT_DIR / "primertrim.amplicon.bed", "w") as outfile:
        outfile.write(panel.to_amplicons(trim_primers=True))

    # Write all the consensus sequences to a single file
    with open(OUTPUT_DIR / "reference.fasta", "w") as reference_outfile:
        reference_records = []
        for msa_obj in msa_dict.values():
            if config.mapping == MappingType.FIRST:
                seq_str = generate_reference(msa_obj.array)
            elif config.mapping == MappingType.CONSENSUS:
                seq_str = generate_consensus(msa_obj.array)
            else:
                raise ValueError("Mapping must be 'first' or 'consensus'")

            reference_records.append(
                SeqRecord.SeqRecord(
                    seq=Seq.Seq(seq_str),
                    id=msa_obj._chrom_name,
                )
            )

        SeqIO.write(reference_records, reference_outfile, "fasta")

    # Generate all the hashes
    ## Generate the bedfile hash, and add it into the config
    primer_md5 = hashlib.md5("\n".join(primer_bed_str).encode()).hexdigest()
    config_dict["primer.bed.md5"] = primer_md5

    ## Generate the amplicon hash, and add it into the config
    amp_md5 = hashlib.md5(amp_bed_str.encode()).hexdigest()
    config_dict["amplicon.bed.md5"] = amp_md5

    ## Read in the reference file and generate the hash
    with open(OUTPUT_DIR / "reference.fasta") as reference_outfile:
        ref_md5 = hashlib.md5(reference_outfile.read().encode()).hexdigest()
    config_dict["reference.fasta.md5"] = ref_md5

    # Write the config dict to file
    # Add the bedfile to the cfg
    config_dict["region_bedfile"] = str(region_bedfile)
    config_dict["input_bedfile"] = str(input_bedfile)
    with open(OUTPUT_DIR / "config.json", "w") as outfile:
        outfile.write(json.dumps(config_dict, sort_keys=True))

    ## DO THIS LAST AS THIS CAN TAKE A LONG TIME
    # Writing plot data
    plot_data = generate_all_plotdata(
        list(msa_dict.values()),
        OUTPUT_DIR / "work",
        last_pp_added=panel._last_pp_added,
    )

    # Write the plot
    with open(OUTPUT_DIR / "plot.html", "w") as outfile:
        outfile.write(
            generate_all_plots_html(plot_data, OUTPUT_DIR, offline_plots=offline_plots)
        )

    with open(OUTPUT_DIR / "primer.html", "w") as outfile:
        for i, msa_obj in enumerate(msa_dict.values()):
            outfile.write(
                primer_mismatch_heatmap(
                    array=msa_obj.array,
                    seqdict=msa_obj._seq_dict,
                    bedfile=OUTPUT_DIR / "primer.bed",
                    offline_plots=True if offline_plots and i == 0 else False,
                )
            )

    logger.info("Completed Successfully")



import gzip
import json
import pathlib
from itertools import groupby
from operator import itemgetter

import numpy as np

from primalscheme3.core.classes import PrimerPair

# Module imports
from primalscheme3.core.msa import MSA
from primalscheme3.core.seq_functions import entropy_score_array

# Panel imports
from primalscheme3.panel.panel_classes import PanelMSA

# Plot format
# plot1: scheme coverage plot. Can be parsed from the bedfile
# plot2 Base occupancy + genome gc
# plot3: Entropy plot
# plot4: Thermo pass Fkmer and Rkmer plot


### Data Scheme
# {chrom_name: {
#     amplicons: {amplicon_number: {s: start, cs: coverage_start, ce: coverage_end, e: end, p: pool, n: name}},
#     dims: [rows, columns],
#     entropy: {position: entropy},
#     gc: {position: gc},
#     occupancy: {position: occupancy},
#     regions: [{s: start, e: end, n: name, sc: score}]
#     thermo_pass: {F: {position: count}, R: {position: count}},
#     uncovered: {start: end},
#     }


def calc_occupancy(align_array: np.ndarray) -> list[tuple[int, float]]:
    results = []
    # Calculate the base proportions
    for index, column in enumerate(align_array.T):
        gaps = np.count_nonzero(column == "-")
        gaps += np.count_nonzero(column == "")
        results.append((index, 1 - (gaps / len(column))))
    return reduce_data(results)


def calc_gc(align_array: np.ndarray, kmer_size: int = 30) -> list[tuple[int, float]]:
    results = []
    # Calculate the base proportions
    for col_index in range(0, align_array.shape[1] - kmer_size, kmer_size):
        slice = align_array[:, col_index : col_index + kmer_size]
        value, counts = np.unique(slice, return_counts=True)

        score_dict = dict(zip(value, counts))
        num_gc = score_dict.get("G", 1) + score_dict.get("C", 1)
        num_at = score_dict.get("A", 1) + score_dict.get("T", 1)

        gc_prop = round(num_gc / (num_at + num_gc), 2)
        results.append((col_index, gc_prop))
    return reduce_data(results)


def reduce_data(results: list[tuple[int, float]]) -> list[tuple[int, float]]:
    """
    Reduce the size of data by merging consecutive points, and rounding to 4 decimal places
    """
    reduced_results = []
    for iindex, (index, oc) in enumerate(results):
        # Add first point
        if iindex == 0:
            reduced_results.append((index, round(oc, 4)))
            continue
        # Add the last point
        if iindex == len(results) - 1:
            reduced_results.append((index, round(oc, 4)))
            continue

        # If the previous point is the same, and the next point is the same
        if results[iindex - 1][1] == oc and results[iindex + 1][1] == oc:
            continue
        else:
            reduced_results.append((index, round(oc, 4)))
    return reduced_results


def generate_uncovered_data(length, primerpairs: list[PrimerPair]) -> dict[int, int]:
    # Set all indexes to uncovered
    uncovered_indexes = {x for x in range(0, length)}

    for primerpair in primerpairs:
        # Handle circular primerpairs
        if primerpair.fprimer.end > primerpair.rprimer.start:
            uncovered_indexes -= set(range(primerpair.fprimer.end, length))
            uncovered_indexes -= set(range(0, primerpair.rprimer.start))
        # Handle linear primerpairs
        else:
            uncovered_indexes -= set(
                range(primerpair.fprimer.end, primerpair.rprimer.start)
            )

    # Plot the uncovered regions
    uncovered_indexes_list = sorted(uncovered_indexes)
    # Generate continous regions
    uncovered_regions = []
    for _k, g in groupby(enumerate(uncovered_indexes_list), lambda ix: ix[0] - ix[1]):
        uncovered_regions.append(list(map(itemgetter(1), g)))

    data: dict[int, int] = dict()
    uncovered_regions = [(min(x), max(x)) for x in uncovered_regions]

    for start, end in uncovered_regions:
        data[start] = end
    return data


def generate_genome_gc_data(msa: MSA | PanelMSA, kmersize: int) -> dict[int, float]:
    """Creates a dict of the genome GC% with key as position and value as GC%"""
    gc_data = dict()
    for index, gc in calc_gc(msa.array, kmersize):
        gc_data[index] = gc
    return gc_data


def generate_genome_occupancy_data(msa: MSA | PanelMSA) -> dict[int, float]:
    """Creates a dict of the genome occupancy with key as position and value as occupancy"""
    occupancy_data = dict()
    for x, y in calc_occupancy(msa.array):
        occupancy_data[x] = y
    return occupancy_data


def generate_genome_entropy_data(msa: MSA | PanelMSA) -> dict[int, float]:
    """Creates a dict of the genome entropy with key as position and value as entropy"""
    results = []
    entropy_data = dict()

    try:
        entropy_array = list(msa._entropy_array)  # type: ignore
    except AttributeError:
        entropy_array = entropy_score_array(msa.array)

    # Calculate the entropy score for each position
    for x, y in enumerate(entropy_array):
        results.append((x, y))
    # Reduce the data
    reduced_data = reduce_data(results)
    for x, y in reduced_data:
        entropy_data[x] = y
    return entropy_data


def generate_thermo_pass_primer_data(msa: MSA | PanelMSA) -> dict[int, str]:
    primer_data = dict()

    fprimer_data = dict()
    for fkmer in msa.fkmers:
        fprimer_data[fkmer.end] = len(fkmer.seqs)
    primer_data["F"] = fprimer_data
    rprimer_data = dict()
    for rkmer in msa.rkmers:
        rprimer_data[rkmer.start] = len(rkmer.seqs)
    primer_data["R"] = rprimer_data
    return primer_data


def generate_amplicon_data(
    primerpairs: list[PrimerPair],
) -> dict[str, dict[str, int | str]]:
    """
    Creates the amplicon plot data
    :param primerpairs: list of PrimerPair objects
    :return: dict of amplicon data
    """
    amplicon_data = dict()

    for primerpair in primerpairs:
        amplicon_data[primerpair.amplicon_number] = {
            "s": min(primerpair.fprimer.starts()),
            "cs": primerpair.fprimer.end,
            "ce": primerpair.rprimer.start,
            "e": max(primerpair.rprimer.ends()),
            "p": primerpair.pool + 1,
            "n": f"{primerpair.amplicon_prefix}_{primerpair.amplicon_number}",
        }

    return amplicon_data


def generate_region_data(msa: MSA | PanelMSA) -> list | None:
    try:
        regions = msa.regions  # type: ignore
    except AttributeError:
        return None

    if not regions:
        return None

    #
    all_data = []
    for region in regions:
        if region.stop >= len(msa._mapping_array):
            end = len(msa._mapping_array) - 1
        else:
            end = msa._ref_to_msa[region.stop]

        all_data.append(
            {
                "s": msa._ref_to_msa[region.start],
                "e": end,
                "n": region.name,
                "sc": region.score,
            }
        )

    return all_data


def generate_data(msa: MSA | PanelMSA, last_pp_added: list[PrimerPair]) -> dict:
    """
    Generate all the plot data for a single MSA
    :param msa: MSA object
    :param pools: The pools object
    :return: dict of all the plot data
    """
    # Filter the last primerpair added to the multiplex
    msa_pp: list[PrimerPair] = [
        x for x in last_pp_added if x.msa_index == msa.msa_index
    ]

    # Remap the included primers to the MSA if they have been mapped to an genome
    if msa._mapping_array is not None:
        for fkmer in msa.fkmers:
            fkmer.end = msa._ref_to_msa[fkmer.end]
            fkmer._starts = {fkmer.end - len(x) for x in fkmer.seqs}
        for rkmer in msa.rkmers:
            rkmer.start = msa._ref_to_msa[rkmer.start]
            rkmer._ends = {rkmer.start + len(x) for x in rkmer.seqs}

    # Write all data to a single json file
    data = dict()
    data["gc"] = generate_genome_gc_data(msa, 30)
    data["entropy"] = generate_genome_entropy_data(msa)
    data["occupancy"] = generate_genome_occupancy_data(msa)
    data["thermo_pass"] = generate_thermo_pass_primer_data(msa)
    data["amplicons"] = generate_amplicon_data(msa_pp)
    data["dims"] = [x for x in msa.array.shape]
    data["uncovered"] = generate_uncovered_data(msa.array.shape[1], msa_pp)
    # Add the region data if pos
    region_data = generate_region_data(msa)
    if region_data is not None:
        data["regions"] = region_data

    return data


def generate_all_plotdata(
    msas: list[MSA] | list[PanelMSA],
    output_path: pathlib.Path,
    last_pp_added: list[PrimerPair],
) -> dict:
    """
    Generate all the plot data for all MSAs to plotdata.json.gz
    :param msa: list of MSA objects
    :param last_pp_added: list of PrimerPair objects added to the multiplex
    :param output_path: pathlib.Path to write the plotdata.json to
    :return: None
    """
    # Write all data to a single json file
    data = dict()
    for msa in msas:
        data[msa._chrom_name] = generate_data(msa, last_pp_added)

    # Write the data to a json file
    json_bytes = json.dumps(data, sort_keys=True, separators=(",", ":")).encode("utf-8")
    with gzip.open(output_path / "plotdata.json.gz", "wb") as fout:
        fout.write(json_bytes)

    # Return the data for use in plotting
    return data



def setup_rich_logger(logfile: str | None = None):
    import logging

    from rich.console import Console
    from rich.logging import RichHandler

    handlers = [RichHandler(level=logging.INFO, markup=True, show_path=False)]
    # Add an optional file handler
    if logfile is not None:
        handlers.append(
            RichHandler(
                level=logging.DEBUG,
                console=Console(file=open(logfile, "w")),
                markup=True,
                show_path=False,
                omit_repeated_times=False,
            )
        )

    logging.basicConfig(
        level="NOTSET", format="%(message)s", datefmt="[%X]", handlers=handlers
    )
    log = logging.getLogger(f"rich-{hash(logfile)}")

    return log



import pathlib

import plotly.graph_objects as go
from plotly.subplots import make_subplots


def generate_all_plots_html(
    plot_data: dict, outdir: pathlib.Path, offline_plots: bool = True
) -> str:
    """Generate all the plots for a scheme from the plot_data"""
    # Generate the plot for each MSA
    plot_html = []

    for index, (chromname, data) in enumerate(plot_data.items()):
        plot_html.append(
            generate_plot_html(
                chromname=chromname,
                msa_data=data,
                outdir=outdir,
                offline_plots=True if offline_plots and index == 0 else False,
            )
        )
    return "\n".join(plot_html)


def generate_plot_html(
    chromname: str, msa_data: dict, outdir: pathlib.Path, offline_plots=True
) -> str:
    # Create an empty figure with the Fprimer hovers
    fig = make_subplots(
        cols=1,
        rows=4,
        shared_xaxes=True,
        vertical_spacing=0.02,
        row_heights=[0.2, 0.2, 0.2, 0.1],
        specs=[
            [{"secondary_y": False}],
            [{"secondary_y": True}],
            [{"secondary_y": False}],
            [{"secondary_y": False}],
        ],
    )

    # Extract amplicon data from the msa_data
    # Filter primers that are circular
    circular_pp = []
    amplicons = []
    for _, pp in msa_data["amplicons"].items():
        if pp["cs"] > pp["ce"]:
            circular_pp.append(pp)
        else:
            amplicons.append(pp)

    length = msa_data["dims"][1]

    if len(amplicons) == 0:
        npools = 0
    else:
        npools = max([x["p"] for x in amplicons])

    fig.add_trace(
        go.Scattergl(
            x=[x["cs"] for x in amplicons],
            y=[x["p"] for x in amplicons],
            opacity=0,
            name="FPrimers",
            hovertext=[f"{x['n']}_LEFT" for x in amplicons],
        ),
        row=1,
        col=1,
    )
    fig.add_trace(
        go.Scattergl(
            x=[x["ce"] for x in amplicons],
            y=[x["p"] for x in amplicons],
            opacity=0,
            name="RPrimers",
            hovertext=[f"{x['n']}_RIGHT" for x in amplicons],
        ),
        row=1,
        col=1,
    )
    # Add the uncovered regions
    for start, stop in msa_data["uncovered"].items():
        fig.add_vrect(
            x0=start,
            x1=stop + 1,
            fillcolor="#F0605D",
            line=dict(width=0),
            opacity=0.5,
            row=1,  # type: ignore
            col=1,  # type: ignore
        )

    # Plot the amplicons lines
    for amplicon in amplicons:
        fig.add_shape(
            type="line",
            y0=amplicon["p"],
            y1=amplicon["p"],
            x0=amplicon["cs"],
            x1=amplicon["ce"],
            line=dict(color="LightSeaGreen", width=5),
            row=1,
            col=1,
        )
        fig.add_shape(
            type="rect",
            y0=amplicon["p"] - 0.05,
            y1=amplicon["p"] + 0.05,
            x0=amplicon["s"],
            x1=amplicon["cs"],
            fillcolor="LightSalmon",
            line=dict(color="darksalmon", width=2),
            row=1,
            col=1,
        )
        fig.add_shape(
            type="rect",
            y0=amplicon["p"] - 0.05,
            y1=amplicon["p"] + 0.05,
            x0=amplicon["ce"],
            x1=amplicon["e"],
            fillcolor="LightSalmon",
            line=dict(color="darksalmon", width=2),
            row=1,
            col=1,
        )

    # Plot the circular primers
    for pp in circular_pp:
        # Add the left side line
        fig.add_shape(
            type="line",
            y0=pp["p"],
            y1=pp["p"],
            x1=pp["ce"],
            x0=0,
            line=dict(color="LightSeaGreen", width=5),
            row=1,
            col=1,
        )
        # Add the right side line
        fig.add_shape(
            type="line",
            y0=pp["p"],
            y1=pp["p"],
            x0=pp["cs"],
            x1=length,
            line=dict(color="LightSeaGreen", width=5),
            row=1,
            col=1,
        )
        fig.add_shape(
            type="rect",
            y0=pp["p"] - 0.05,
            y1=pp["p"] + 0.05,
            x0=pp["s"],
            x1=pp["cs"],
            fillcolor="LightSalmon",
            line=dict(color="LightSalmon", width=2),
            row=1,
            col=1,
        )
        fig.add_shape(
            type="rect",
            y0=pp["p"] - 0.05,
            y1=pp["p"] + 0.05,
            x0=pp["ce"],
            x1=pp["e"],
            fillcolor="LightSalmon",
            line=dict(color="LightSalmon", width=2),
            row=1,
            col=1,
        )

    # Add the base occupancy
    occupancy_data = [
        (int(index), float(oc)) for index, oc in msa_data["occupancy"].items()
    ]
    fig.add_trace(
        go.Scattergl(
            x=[x[0] for x in occupancy_data],
            y=[x[1] for x in occupancy_data],
            mode="lines",
            name="Base Occupancy",
            line=dict(color="#F0605D", width=2),
            fill="tozeroy",
            opacity=0.5,
        ),
        row=2,
        col=1,
    )

    ## Plot the GC data
    gc_data = [(int(index), float(gc)) for index, gc in msa_data["gc"].items()]
    fig.add_trace(
        go.Scattergl(
            x=[x[0] for x in gc_data],
            y=[x[1] for x in gc_data],
            mode="lines",
            name="GC Prop",
            line=dict(color="#005c68", width=2),
        ),
        row=2,
        col=1,
        secondary_y=True,
    )

    # Add the entropy plot
    entropy_data = [
        (int(index), float(entropy)) for index, entropy in msa_data["entropy"].items()
    ]
    fig.add_trace(
        go.Scattergl(
            x=[x[0] for x in entropy_data],
            y=[x[1] for x in entropy_data],
            opacity=1,
            name="Sequence Entropy",
            mode="lines",
        ),
        row=3,
        col=1,
    )

    # Add all possible Fkmers
    fkmer_data = [
        (end, num_seqs) for end, num_seqs in msa_data["thermo_pass"]["F"].items()
    ]
    fig.add_trace(
        go.Scattergl(
            x=[x[0] for x in fkmer_data],
            y=[1 for _ in fkmer_data],
            hovertext=[x[1] for x in fkmer_data],
            marker=dict(symbol="triangle-right", size=10),
            mode="markers",
            name="Passing Forward Primers",
            hovertemplate="Number Seqs: %{hovertext}",
        ),
        row=4,
        col=1,
    )
    # Add all possible Rkmers
    rkmer_data = [
        (start, num_seqs) for start, num_seqs in msa_data["thermo_pass"]["R"].items()
    ]
    fig.add_trace(
        go.Scattergl(
            x=[x[0] for x in rkmer_data],
            y=[0.5 for _ in rkmer_data],
            hovertext=[x[1] for x in rkmer_data],
            marker=dict(symbol="triangle-left", size=10),
            mode="markers",
            name="Passing Reverse Primers",
            hovertemplate="Number Seqs: %{hovertext}",
        ),
        row=4,
        col=1,
    )

    # Add the regions
    # If regions are present add them
    if "regions" in msa_data:
        data = msa_data.get("regions", [])
        for region in data:
            fig.add_shape(
                type="line",
                y0=1.5,
                y1=1.5,
                x0=region["s"],
                x1=region["e"],
                fillcolor="Green",
                line=dict(color="Green", width=5),
                row=1,
                col=1,
            )
        fig.add_trace(
            go.Scatter(  # doesn't need WebGL
                x=[x["s"] + ((x["e"] - x["s"]) // 2) for x in data],
                y=[1.5 for _ in data],
                name="Regions",
                mode="markers",
                text=[x["n"] for x in data],
                hovertemplate="%{text}",
                opacity=0,  # Make the markers invisible
            ),
        )

    # Add the base plot settings
    fig.update_xaxes(
        showline=True,
        mirror=True,
        ticks="outside",
        linewidth=2,
        linecolor="black",
        tickformat=",d",
        title_font=dict(size=18, family="Arial", color="Black"),
        range=[0, length],
        title="",  # Blank title for all x-axes
    )
    fig.update_yaxes(
        showline=True,
        mirror=True,
        ticks="outside",
        linewidth=2,
        linecolor="black",
        fixedrange=True,
        title_font=dict(size=18, family="Arial", color="Black"),
    )

    # Update the top plot
    fig.update_yaxes(
        range=[0.5, npools + 0.5],
        title="pool",
        tickmode="array",
        tickvals=sorted({x["p"] for x in amplicons}),
        row=1,
        col=1,
    )
    # Update the second plot
    fig.update_yaxes(
        range=[-0.1, 1.1],
        title="Base Occupancy",
        tickmode="array",
        tickvals=[0, 0.25, 0.5, 0.75, 1],
        row=2,
        col=1,
        secondary_y=False,
    )
    fig.update_yaxes(
        range=[-0.1, 1.1],
        title="GC%",
        tickmode="array",
        tickvals=[0, 0.25, 0.5, 0.75, 1],
        ticktext=[0, 25, 50, 75, 100],
        row=2,
        col=1,
        secondary_y=True,
        side="right",
    )
    # Update the third plot
    fig.update_yaxes(
        title="Entropy",
        tickmode="array",
        row=3,
        col=1,
    )
    # Update the fourth plot
    fig.update_yaxes(
        title="Thermo-passing Primers",
        range=[0.5 - 0.1, 1 + 0.1],
        tickmode="array",
        row=4,
        col=1,
        secondary_y=False,
    )
    fig.update_xaxes(
        title="Position", row=4, col=1
    )  # Add the x-axis title to the bottom plot

    # fig.update_layout(paper_bgcolor="#000000")
    fig.update_layout(height=900, title_text=chromname, showlegend=False)
    # plot_bgcolor="rgba(246, 237, 202, 0.5)",

    # Remove unnecessary plot elements
    fig.update_layout(
        modebar_remove=[
            "select2d",
            "lasso2d",
            "select",
            "autoScale2d",
            "zoom",
            "toImage",
        ]
    )

    # Write a png version of the plot
    fig.write_image(
        str(outdir.absolute() / (chromname + ".png")),
        format="png",
        height=900,
        width=1600,
    )

    # Write a html version of the plot
    return fig.to_html(
        include_plotlyjs=True if offline_plots else "cdn", full_html=False
    )



import logging
import pathlib
import re
import sys

from primalscheme3.core.classes import FKmer, PrimerPair, RKmer
from primalscheme3.core.config import Config

# Module imports
from primalscheme3.core.seq_functions import expand_ambs

REGEX_PATTERN_PRIMERNAME = re.compile("\\d+(_RIGHT|_LEFT|_R|_L)")


def re_primer_name(string) -> list[str] | None:
    """
    Will return (amplicon_number, R/L) or None
    """
    match = REGEX_PATTERN_PRIMERNAME.search(string)
    if match:
        return match.group().split("_")
    return None


class BedPrimerPair(PrimerPair):
    """Class to contain a single primercloud from a bedfile, which contains the extra info parsed from the bedfile"""

    amplicon_prefix: str
    # Calc values
    _primername: str

    def __init__(
        self,
        fprimer: FKmer,
        rprimer: RKmer,
        msa_index: int,
        chrom_name: str,
        amplicon_prefix: str,
        amplicon_number: int,
        pool: int,
    ) -> None:
        self.fprimer = fprimer
        self.rprimer = rprimer
        self.chrom_name = chrom_name
        self.amplicon_prefix = amplicon_prefix
        self.msa_index = msa_index
        self.amplicon_number = amplicon_number
        self.pool = pool

        #
        self._primername = f"{self.amplicon_number}_{self.amplicon_prefix}"

    def match_primer_stem(self, primernamestem: str) -> bool:
        return self._primername == primernamestem


class BedLine:
    """
    Contains a single line from a bedfile
    self.pool is stored as a 0 based index
    """

    chrom_name: str
    _start: int
    _end: int
    primername: str
    pool: int
    direction: str
    sequence: str
    # Calc values
    amplicon_number: int

    def __init__(self, bedline: list[str]) -> None:
        self.chrom_name = bedline[0]
        self._start = int(bedline[1])
        self._end = int(bedline[2])
        self.primername = bedline[3]
        self.pool = int(bedline[4]) - 1
        self.direction = bedline[5]
        self.sequence = bedline[6]

        # Check primername is valid
        if len(self.primername.split("_")) != 4:
            raise ValueError(
                f"Invalid primername: {self.primername} in bedline: {bedline}"
            )

        # Calc values
        self.amplicon_number = int(self.primername.split("_")[1])

    def all_seqs(self) -> set[str] | None:
        "Expands ambs bases"
        return expand_ambs([self.sequence])

    @property
    def msa_index(self) -> str:
        return self.chrom_name

    @property
    def start(self) -> int:
        return self._start

    @property
    def end(self) -> int:
        return self._end

    def __str__(self, *kwargs) -> str:
        # I use *kwargs so that it can have the same behaviour as PrimerPairs
        return f"{self.chrom_name}\t{self.start}\t{self.end}\t{self.primername}\t{self.pool + 1}\t{self.direction}\t{self.sequence}"


def read_in_bedlines(path: pathlib.Path) -> tuple[list[BedLine], list[str]]:
    """
    Read in bedlines from a file.

    :param path: The path to the bed file.
    :type path: pathlib.Path
    :return: A list of BedLine objects.
    :rtype: tuple(list[BedLine], list[str])
    """
    bed_primers: list[BedLine] = []
    bed_headers: list[str] = []
    with open(path) as bedfile:
        for line in bedfile.readlines():
            line = line.strip()
            if not line:  # Skip empty lines
                continue
            elif line.startswith("#"):  # Store header lines
                bed_headers.append(line.strip())
            else:  # Store primer lines
                line = line.strip().split()
                bed_primers.append(BedLine(line))
    return (bed_primers, bed_headers)


def read_in_bedprimerpairs(path: pathlib.Path) -> tuple[list[BedPrimerPair], list[str]]:
    """
    Read in a bedfile and return a list of BedPrimerPairs, MSA index is set to None as it is not known at this point

    :param path: The path to the bed file.
    :type path: pathlib.Path
    :return: A list of BedPrimerPair objects, and the header lines from the bedfile.
    :rtype: tuple(list[BedPrimerPair], list[str])
    """

    # Read in the bedfile
    primerpairs = []
    primerlines, _headers = read_in_bedlines(path)  # Ignore headers for now

    # Group primers by reference
    ref_to_bedlines: dict[str, list[BedLine]] = dict()
    for ref in {bedline.chrom_name for bedline in primerlines}:
        ref_to_bedlines[ref] = [x for x in primerlines if x.chrom_name == ref]

    for ref, ref_bed_lines in ref_to_bedlines.items():
        # Group the bedlines by amplicon number
        for ampliconnumber in {
            int(bedline.amplicon_number) for bedline in ref_bed_lines
        }:
            amplicon_prefix = ref_bed_lines[0].primername.split("_")[0]
            ampliconlines = [
                x for x in ref_bed_lines if x.amplicon_number == ampliconnumber
            ]
            pool = ampliconlines[0].pool

            fp = [x for x in ampliconlines if x.direction == "+"]
            rp = [x for x in ampliconlines if x.direction == "-"]

            if len(fp) == 0:
                raise ValueError(
                    f"Primer {ampliconlines[0].primername} has no forward primer"
                )
            if len(rp) == 0:
                raise ValueError(
                    f"Primer {ampliconlines[0].primername} has no reverse primer"
                )
            # Group the ampliconlines by direction
            fkmer = FKmer(
                max([x.end for x in fp]),
                [x.sequence for x in fp],
            )
            rkmer = RKmer(
                min([x.start for x in rp]),
                [x.sequence for x in rp],
            )
            primerpairs.append(
                BedPrimerPair(
                    fprimer=fkmer,
                    rprimer=rkmer,
                    msa_index=None,  # This is set later # type: ignore
                    chrom_name=ref,
                    amplicon_number=int(ampliconnumber),
                    amplicon_prefix=amplicon_prefix,
                    pool=pool,
                )
            )

    primerpairs.sort(key=lambda x: (x.chrom_name, x.amplicon_number))
    return (primerpairs, _headers)


def create_bedfile_str(
    headers: list[str] | None, primerpairs: list[PrimerPair | BedPrimerPair]
) -> str:
    """
    Returns the multiplex as a bed file
    :return: str
    """
    primer_bed_str: list[str] = []

    # Ensure headers are commented and valid
    if headers is not None:
        for headerline in headers:
            if not headerline.startswith("#"):
                headerline = "# " + headerline
            primer_bed_str.append(headerline.strip())

    # Add the primerpairs to the bed file
    for pp in primerpairs:
        primer_bed_str.append(pp.to_bed().strip())

    return "\n".join(primer_bed_str) + "\n"


def create_amplicon_str(
    primerpairs: list[PrimerPair | BedPrimerPair], trim_primers: bool = False
) -> str:
    amplicon_str: list[str] = []
    # Add the amplicons to the string
    for pp in primerpairs:
        if trim_primers:
            amplicon_str.append(
                f"{pp.chrom_name}\t{pp.fprimer.region()[1]}\t{pp.rprimer.region()[0] - 1}\t{pp.amplicon_prefix}_{pp.amplicon_number}\t{pp.pool + 1}"
            )
        else:
            amplicon_str.append(
                f"{pp.chrom_name}\t{pp.fprimer.region()[0]}\t{pp.rprimer.region()[1]}\t{pp.amplicon_prefix}_{pp.amplicon_number}\t{pp.pool + 1}"
            )
    return "\n".join(amplicon_str) + "\n"


def read_in_extra_primers(
    input_bedfile: pathlib.Path, config: Config, logger: logging.Logger
) -> list[BedPrimerPair]:
    """
    Reads in Primers from a bedfile, and QC checks them for Tm and Pools
    """
    bedprimerpairs, _headers = read_in_bedprimerpairs(input_bedfile)

    logger.info(
        f"Read in bedfile: [blue]{input_bedfile.name}[/blue]: "
        f"[green]{len(bedprimerpairs)}[/green] PrimersPairs containing "
        f"{len([primer for primers in (bedprimerpair.all_seqs() for bedprimerpair in bedprimerpairs) for primer in primers])} primers",
    )

    # Check the primers for Tm
    primer_tms = [
        tm for tm in (pp.calc_tm(config) for pp in bedprimerpairs) for tm in tm
    ]
    if min(primer_tms) < config.primer_tm_min or max(primer_tms) > config.primer_tm_max:
        logger.warning(
            f"Primer Tm outside range: {round(min(primer_tms), 2)} : {round(max(primer_tms),2)} (range: {config.primer_tm_min} : {config.primer_tm_max})"
        )

    else:
        logger.info(
            f"Primer Tm range: [green]{min(primer_tms)}[/green] : [green]{max(primer_tms)}[/green]"
        )

    # Check pools are within the range
    pools_in_bed = {primer.pool for primer in bedprimerpairs}
    if max(pools_in_bed) > config.n_pools:
        logger.critical(
            f"The number of pools in the bedfile is greater than --npools: "
            f"{max(pools_in_bed)} > {config.n_pools}"
        )
        sys.exit(1)

    return bedprimerpairs



# This script contains tools to polish the output of the primal scheme.

from primaldimer_py import which_kmers_pools_interact  # type: ignore

from primalscheme3.core.classes import FKmer, PrimerPair, RKmer


def find_interactions(
    pool: list[PrimerPair], primerpair: PrimerPair, threshold: float, calc_all=True
) -> list[PrimerPair]:
    """Find all interactions between pools"""
    interacting_kmers: list[tuple[FKmer | RKmer, FKmer | RKmer]] = (
        which_kmers_pools_interact(
            [kmer for kmers in (pp.kmers() for pp in pool) for kmer in kmers],
            [*primerpair.kmers()],
            threshold,
            calc_all,
        )
    )

    # map the interacting kmers to the primer pairs
    # TODO speed up this part
    interacting_primerpairs = []
    for kmer1, kmer2 in interacting_kmers:
        if kmer1 in primerpair.kmers():
            interacting_primerpairs.extend([pp for pp in pool if kmer2 in pp.kmers()])
        elif kmer2 in primerpair.kmers():
            interacting_primerpairs.extend([pp for pp in pool if kmer1 in pp.kmers()])
    return interacting_primerpairs


if __name__ == "__main__":
    fkmer = FKmer(10, ["TGGAAATACCCACAAGTTAATGGTTTAAC"])
    rkmer = RKmer(100, ["ACACCTGTGCCTGTTAAACCAT"])

    pp = PrimerPair(fkmer, rkmer, 0)
    print(
        [kmer for kmers in (pp.kmers() for pp in [pp]) for kmer in kmers],
    )

    interaction = which_kmers_pools_interact([fkmer, rkmer], [*pp.kmers()], -26, True)



from collections.abc import Iterator

from tqdm import tqdm


class ProgressTracker(tqdm):
    def __init__(self, parent, process, iterable, chrom, count=0, *args, **kwargs):
        self.parent = parent
        self.process = process
        self.chrom = chrom
        self.count = count
        super().__init__(iterable, *args, **kwargs)

    def __iter__(self) -> Iterator:
        obj = super().__iter__()

        for i in obj:
            self.parent.signal()
            yield i

    def manual_update(
        self,
        n: int | None = None,
        total: int | None = None,
        process: str | None = None,
        chrom: str | None = None,
        count: int | float | None = None,
        update: bool = False,
    ):
        old_n = self.n
        # Update the progress bar manually
        if n is not None:
            self.n = n
        if total is not None:
            self.total = total
        if process is not None:
            self.process = process
        if chrom is not None:
            self.chrom = chrom
        if count is not None:
            self.count = count

        if update:
            self.update(self.n - old_n)

        self.parent.signal()


class ProgressManager:
    _subprocess: None | ProgressTracker

    def signal(self):
        pass

    def __init__(self):
        self._status = None
        self._subprocess = None

    def n(self) -> int | None:
        if self._subprocess:
            return self._subprocess.n
        return None

    def total(self) -> int | None:
        if self._subprocess:
            return self._subprocess.total
        return None

    def process(self) -> str | None:
        if self._subprocess:
            return self._subprocess.process
        return None

    def chrom(self) -> str | None:
        if self._subprocess:
            return self._subprocess.chrom
        return None

    def count(self) -> int | float | None:
        if self._subprocess:
            return self._subprocess.count
        return None

    def create_sub_progress(
        self, iter, chrom, process, *args, **kwargs
    ) -> ProgressTracker:
        """Create a progress tracker"""
        self._subprocess = ProgressTracker(
            self,
            *args,
            iterable=iter,
            chrom=chrom,
            process=process,
            **kwargs,
            desc=process,
        )
        return self._subprocess



from enum import Enum

import numpy as np
from primaldimer_py import do_pools_interact_py  # type: ignore

from primalscheme3.core.bedfiles import (
    BedPrimerPair,
    create_amplicon_str,
    create_bedfile_str,
)
from primalscheme3.core.classes import MatchDB, PrimerPair
from primalscheme3.core.config import Config
from primalscheme3.core.mismatches import detect_new_products
from primalscheme3.core.msa import MSA


class PrimerPairCheck(Enum):
    """
    Enum for checking if a primerpair can be added to a pool
    """

    OVERLAP = 1
    INTERACTING = 2
    MISPRIMING = 3
    OK = 4
    CIRCULAR = 5


class Multiplex:
    """
    This is the baseclass for all multiplexes (Scheme / Panel)
    - It allows multiple pools

    Params:
    - cfg: dict. The configuration dict
    - matchDB: MatchDB object
    - msa_dict: dict[int, MSA]. A dict of MSA objects

    Internal:
    - self._pools: list[list[PrimerPair | BedPrimerPair]]. List of pools with PrimerPairs
    - self._current_pool: int. The current pool number
    - self._last_pp_added: list[PrimerPair]. Stack to keep track of the last primer added
    - self._matchDB: MatchDB object
    - self._matches: list[set[tuple]]. A list of sets with matches for each pool
    - self._msa_dict: dict[int, MSA]. A dict of MSA objects
    - self._coverage: dict[int, np.ndarray]. PrimerTrimmed Coverage for each MSA index
    - self._lookup: dict[int, np.ndarray[npools, ncols]] | None. A lookup for the primerpairs in the multiplex.
    """

    _pools: list[list[PrimerPair | BedPrimerPair]]
    _current_pool: int
    _last_pp_added: list[PrimerPair]  # Stack to keep track of the last primer added
    _matchDB: MatchDB
    _matches: list[set[tuple]]
    _msa_dict: dict[int, MSA]
    _coverage: dict[int, np.ndarray]  # PrimerTrimmed Coverage for each MSA index
    _lookup: dict[int, np.ndarray]
    config: Config

    def __init__(
        self, config: Config, matchDB: MatchDB, msa_dict: dict[int, MSA]
    ) -> None:
        self.n_pools = config.n_pools
        self._pools = [[] for _ in range(self.n_pools)]
        self._matches: list[set[tuple]] = [set() for _ in range(self.n_pools)]
        self._current_pool = 0
        self._pp_number = 1
        self.config = config
        self._matchDB = matchDB
        self._last_pp_added = []
        self._msa_dict = msa_dict
        # Set up coverage dict
        self.setup_coverage()

        # Set up the lookup
        self.setup_primerpair_lookup()

    def setup_primerpair_lookup(self) -> None:
        """
        Returns a lookup of primerpairs in the multiplex
        :return: None
        """
        self._lookup = {}
        for msa_index, msa in self._msa_dict.items():
            if msa._mapping_array is None:
                n = len(msa.array[1])
            else:
                n = len(msa._mapping_array)

            n = [None] * n
            # Create a lookup for the primerpairs
            self._lookup[msa_index] = np.array(
                [n for _ in range(self.n_pools)], ndmin=2
            )

    def update_lookup(self, primerpair: PrimerPair | BedPrimerPair, add: bool) -> None:
        """
        Updates the lookup for the primerpair
        :param primerpair: PrimerPair object
        :param add: bool. If True, add to the lookup. If False, remove from the lookup
        :return: None
        """
        circular = primerpair.fprimer.end > primerpair.rprimer.start

        # If the msa_index is not in the lookup. Then return
        if primerpair.msa_index not in self._lookup:
            return

        value = primerpair if add else None

        if circular:
            self._lookup[primerpair.msa_index][
                primerpair.pool, primerpair.fprimer.region()[0] :
            ] = value
            self._lookup[primerpair.msa_index][
                primerpair.pool, : primerpair.rprimer.region()[1]  # slice non inclusive
            ] = value
        else:
            self._lookup[primerpair.msa_index][
                primerpair.pool,
                primerpair.fprimer.region()[0] : primerpair.rprimer.region()[
                    1
                ],  # slice non inclusive
            ] = value

    def get_seqs_in_pool(self, pool: int) -> list[str]:
        """
        Returns a list of all the sequences in the pool
        :param pool: int
        :return: list[str]
        """
        return [
            seq
            for ppseq in (pp.all_seqs() for pp in self._pools[pool])
            for seq in ppseq
        ]

    def check_primerpair_can_be_added(
        self, primerpair: PrimerPair, pool: int, otherseqs: list[str] | None = None
    ) -> PrimerPairCheck:
        """
        Checks if the primerpair can be added to the multiplex
        - Checks if the primerpair overlaps with any other primerpairs in the pool
        - Checks if the primerpair has any matches in the MatchDB
        :param primerpair: PrimerPair object
        :param otherseqs: list[str] | None. A list of other sequences to check for interactions against
        :return: bool. True if the primerpair can be added
        """
        # Check if the primerpair overlaps with any other primerpairs in the pool
        if self.does_overlap(primerpair, pool):
            return PrimerPairCheck.OVERLAP

        # Check for interactions with other sequences
        if otherseqs is None:
            otherseqs = self.get_seqs_in_pool(pool)
        if do_pools_interact_py(
            primerpair.all_seqs(),
            otherseqs,
            self.config.dimer_score,
        ):
            return PrimerPairCheck.INTERACTING

        # Check for other PCR Products
        if detect_new_products(
            primerpair.find_matches(
                self._matchDB,
                remove_expected=False,
                kmersize=self.config.mismatch_kmersize,
                fuzzy=self.config.mismatch_fuzzy,
            ),
            self._matches[pool],
            self.config.mismatch_product_size,
        ):
            return PrimerPairCheck.MISPRIMING

        # If all checks pass, return True
        return PrimerPairCheck.OK

    def setup_coverage(self) -> None:
        """
        Sets up the coverage dict
        :param msa_dict: dict[int, MSA]
        :return: None
        """
        self._coverage = {}
        for msa_index, msa in self._msa_dict.items():
            if msa._mapping_array is None:
                n = len(msa.array[1])
            else:
                n = len(msa._mapping_array)
            self._coverage[msa_index] = np.array([False] * n)

    def recalculate_coverage(self) -> None:
        """
        Recalculates the primertrimmed coverage for all MSA indexes
        :param msa_dict: dict[int, MSA]
        :return: None
        """
        # Create an empty coverage dict
        self.setup_coverage()
        for pp in self.all_primerpairs():
            self.update_coverage(pp, add=True)

    def get_coverage_percent(self, msa_index: int) -> float:
        """
        Returns the coverage percentage for the specified MSA index
        :param msa_index: int
        :return: float | None
        """
        return round(
            self._coverage[msa_index].sum() / len(self._coverage[msa_index]) * 100, 2
        )

    def get_coverage_gaps(self, msa_index: int) -> list[tuple[int, int]]:
        """
        Returns the coverage gaps for the specified MSA index
        :param msa_index: int
        :return: list[tuple[int, int]]
        """
        gaps = []
        in_gap = False
        gap_start = 0
        for i, covered in enumerate(self._coverage[msa_index]):
            if not covered:
                if not in_gap:
                    gap_start = i
                    in_gap = True
            else:
                if in_gap:
                    gaps.append((gap_start, i))
                    in_gap = False
        if in_gap:
            gaps.append((gap_start, i))  # type: ignore
        return gaps

    def update_coverage(self, primerpair: PrimerPair, add: bool = True) -> None:
        """
        Updates the coverage for the specified MSA index
        :param msa_index: int
        :param primerpair: PrimerPair object
        :param add: bool. If True, add to the coverage. If False, remove from the coverage
        :return: None
        """
        # If the msa_index is not in the lookup. Then return
        if primerpair.msa_index not in self._coverage:
            return
        # Check circular
        if primerpair.fprimer.end > primerpair.rprimer.start:
            # Handle circular
            self._coverage[primerpair.msa_index][primerpair.fprimer.region()[1] :] = add
            self._coverage[primerpair.msa_index][: primerpair.rprimer.region()[0]] = (
                add  # slice non inclusive
            )
        else:
            self._coverage[primerpair.msa_index][
                primerpair.fprimer.end : primerpair.rprimer.start
            ] = add

    def next_pool(self) -> int:
        """
        Returns the next pool number.
        Does not directly change self._current_pool
        :return: int
        """
        return (self._current_pool + 1) % self.n_pools

    def add_primer_pair_to_pool(
        self, primerpair: PrimerPair | BedPrimerPair, pool: int, msa_index: int
    ):
        """
        Main method to add a primerpair to a pool. Performs no checks.
        - Adds PrimerPair to the specified pool
        - Updates the PrimerPair's pool and amplicon_number
        - Updates the pools matches
        - Appends PrimerPair to _last_pp_added
        - Sets the Multiplex to the specified pool. Then moves the Multiplex to the next pool


        :param primerpair: PrimerPair object
        :param pool: int
        :param msa_index: int
        :return: None
        """
        # Set the primerpair values
        primerpair.pool = pool
        primerpair.amplicon_number = (
            len(
                [
                    pp
                    for sublist in self._pools
                    for pp in sublist
                    if pp.msa_index == primerpair.msa_index
                ]
            )
            + 1
        )

        # Adds the primerpair's matches to the pools matches
        self._matches[pool].update(
            primerpair.find_matches(
                self._matchDB,
                fuzzy=self.config.mismatch_fuzzy,
                remove_expected=True,
                kmersize=self.config.mismatch_kmersize,
            )
        )

        # Adds the primerpair to the pool
        self._pools[pool].append(primerpair)
        self._current_pool = pool
        self._current_pool = self.next_pool()
        self._last_pp_added.append(primerpair)

        # Update the lookup
        self.update_lookup(primerpair, add=True)
        # Update the coverage
        self.update_coverage(primerpair, add=True)

    def remove_primerpair(self, pp):
        """
        Main method to remove a primerpair from the multiplex
        - Removes the primerpair from the pool
        - Removes the primerpair's matches from the pool's matches
        - Updates the lookup
        - Updates the coverage
        :param pp: PrimerPair object
        :return: None
        """
        # Removes the pp from stack
        self._last_pp_added.remove(pp)

        # Remove the primerpair from the pool
        self._pools[pp.pool].remove(pp)

        # Remove the primerpair's matches from the pool's matches
        self._matches[pp.pool].difference_update(
            pp.find_matches(
                self._matchDB,
                fuzzy=self.config.mismatch_fuzzy,
                remove_expected=False,
                kmersize=self.config.mismatch_kmersize,
            )
        )

        # Update the lookup
        self.update_lookup(pp, add=False)
        # Update the coverage
        self.update_coverage(pp, add=False)

    def remove_last_primer_pair(self) -> PrimerPair:
        """
        This removes the last primerpair added
        - Finds the last primerpair added from self._last_pp_added
        - Removes the primerpair from the pool
        - Removes the primerpair's matches from the pool's matches
        - Moves the current pool to the last primerpair's pool
        - Returns the last primerpair added
        :raises: IndexError if no primerpairs have been added

        :return: PrimerPair object
        """
        last_pp = self._last_pp_added[-1]
        self.remove_primerpair(last_pp)
        self._current_pool = last_pp.pool

        return last_pp

    def does_overlap(self, primerpair: PrimerPair | BedPrimerPair, pool: int) -> bool:
        """
        Does this primerpair overlap with any primerpairs in the pool?
        :param primerpair: PrimerPair object
        :param pool: int
        :return: bool. True if overlaps
        """
        # Circular
        if primerpair.fprimer.end > primerpair.rprimer.start:
            return (
                np.count_nonzero(
                    self._lookup[primerpair.msa_index][
                        pool, primerpair.fprimer.region()[0] :
                    ]
                )
                > 0
                or np.count_nonzero(
                    self._lookup[primerpair.msa_index][
                        pool, : primerpair.rprimer.region()[1]
                    ]
                )
                > 0
            )

        # Get the slice of the lookup
        return (
            np.count_nonzero(
                self._lookup[primerpair.msa_index][
                    pool,
                    primerpair.fprimer.region()[0] : primerpair.rprimer.region()[
                        1
                    ],  # slice non inclusive
                ]
            )
            > 0
        )

    def all_primerpairs(self) -> list[PrimerPair]:
        """
        Returns a list of all primerpairs in the multiplex.
        Sorted by MSA index and amplicon number
        :return: list[PrimerPair]
        """
        all_pp = [pp for pool in (x for x in self._pools) for pp in pool]
        all_pp.sort(key=lambda pp: (str(pp.msa_index), pp.amplicon_number))
        return all_pp

    # Output methods
    def to_bed(
        self,
        headers: list[str] | None = None,
    ) -> str:
        """
        Returns the multiplex as a bed file
        :return: str
        """
        if headers is None:
            headers = ["# artic-bed-version v3.0"]

        return create_bedfile_str(headers, self.all_primerpairs())

    def to_amplicons(
        self,
        trim_primers: bool,
    ) -> str:
        """
        Returns the multiplex as an amplicon file
        :param trim_primers: bool. If True, the primers are trimmed from the amplicons
        :return: str
        """
        return create_amplicon_str(self.all_primerpairs(), trim_primers)

    def polish(
        self,
        msas_dict: dict[int, MSA],
    ) -> None:
        """
        Stochastic optimization to improve the multiplex
        """

        # Create interaction network

        # Find primerpairs that cover uncoverered regions
        msaindex_to_primerpairs = {}
        for msa_index, msa in msas_dict.items():
            for pp in msa.primerpairs:
                numuncoveredpos = (
                    pp.rprimer.start
                    - pp.fprimer.end
                    - np.sum(
                        self._coverage[msa_index][pp.fprimer.end : pp.rprimer.start],
                        dtype=int,
                    )
                )
                if numuncoveredpos > 0:
                    if msa_index not in msaindex_to_primerpairs:
                        msaindex_to_primerpairs[msa_index] = []
                    msaindex_to_primerpairs[msa_index].append((pp, numuncoveredpos))

        pass



import pathlib

import numpy as np
import plotly.graph_objects as go
from click import UsageError
from plotly.subplots import make_subplots

# Create in the classes from primalscheme3
from primalscheme3.core.bedfiles import (
    BedLine,
    read_in_bedlines,
    read_in_bedprimerpairs,
)
from primalscheme3.core.create_report_data import calc_gc
from primalscheme3.core.mapping import (
    check_for_end_on_gap,
    create_mapping,
    fix_end_on_gap,
    ref_index_to_msa,
)
from primalscheme3.core.seq_functions import extend_ambiguous_base, reverse_complement


class PlotlyText:
    """
    A class to hold the text for a plotly heatmap.
    """

    primer_name: str
    primer_seq: str
    genome_seq: str

    def __init__(
        self,
        primer_name: str,
        primer_seq: str,
        genome_seq: str,
    ):
        self.primer_name = primer_name
        self.primer_seq = primer_seq
        self.genome_seq = genome_seq

    def format_str(self) -> str:
        # parsedseqs
        cigar = []
        for p, g in zip(self.primer_seq[::-1], self.genome_seq[::-1]):
            if p == g:
                cigar.append("|")
            else:
                cigar.append(".")
        cigar = "".join(cigar)[::-1]
        return f"5'{self.primer_seq}: {self.primer_name}<br>5'{cigar}<br>5'{self.genome_seq[-len(self.primer_seq):]}"


def get_primers_from_msa(
    array: np.ndarray, index: int, forward: bool = True, length: int = 20, row=0
) -> dict[int, str | None]:
    """
    Get a primer from an MSA array.
    """
    row_data = {}
    if forward:
        for row in range(array.shape[0]):
            gaps = 0
            row_data[row] = None
            while index - length - gaps >= 0:
                # Get slice
                initial_slice = array[row, index - length - gaps : index]
                # Check for gaps on set base
                if initial_slice[-1] == "-":
                    break
                sequence = "".join(initial_slice).replace("-", "")
                # Covered removed gaps
                if "" in initial_slice:
                    break
                # Check for gaps in the slice
                if len(sequence) == length:
                    row_data[row] = sequence
                    break
                # Walk left
                gaps += 1
    else:
        for row in range(array.shape[0]):
            gaps = 0
            row_data[row] = None
            while index + length + gaps <= array.shape[1]:
                # Get slice
                initial_slice = array[row, index : index + length + gaps]
                # Check for gaps on set base
                if initial_slice[0] == "-":
                    break
                sequence = "".join(initial_slice).replace("-", "")
                # Covered removed gaps
                if "" in initial_slice:
                    break
                # Check for gaps in the slice
                if len(sequence) == length:
                    row_data[row] = reverse_complement(sequence)
                    break
                # Walk right
                gaps += 1
    return row_data


def calc_primer_hamming(seq1, seq2) -> int:
    """
    Calculate the hamming distance between two sequences of equal length. Ignores N.
    :param seq1: The primer sequence in 5' to 3' orientation.
    :param seq2: The primer sequence in 5' to 3' orientation.
    :return: The number of mismatches between the two sequences.
    """
    dif = 0
    for seq1b, seq2b in zip(seq1[::-1], seq2[::-1]):
        seq1b_exp = set(extend_ambiguous_base(seq1b))
        seq2b_exp = set(extend_ambiguous_base(seq2b))

        if not seq1b_exp & seq2b_exp and (seq1b != "N" and seq2b != "N"):
            dif += 1

    return dif


def primer_mismatch_heatmap(
    array: np.ndarray,
    seqdict: dict,
    bedfile: pathlib.Path,
    include_seqs: bool = True,
    offline_plots: bool = True,
) -> str:
    """
    Create a heatmap of primer mismatches in an MSA.
    :param array: The MSA array.
    :param seqdict: The sequence dictionary.
    :param bedfile: The bedfile of primers.
    :param include_seqs: Reduces plot size by removing hovertext.
    :raises: click.UsageError
    """
    # Read in the bedfile
    bedlines, _header = read_in_bedlines(bedfile)

    # Find the mapping genome
    bed_chrom_names = {bedline.chrom_name for bedline in bedlines}

    # Reference genome
    primary_ref = bed_chrom_names.intersection(seqdict.keys())

    if len(primary_ref) == 0:
        # Try to fix a common issue with Jalview
        parsed_seqdict = {"_".join(k.split("/")): v for k, v in seqdict.items()}
        primary_ref = bed_chrom_names.intersection(parsed_seqdict.keys())
        seqdict = parsed_seqdict

    # Filter the bedlines for only the reference genome
    bedlines = [bedline for bedline in bedlines if bedline.chrom_name in primary_ref]

    # handle errors
    if len(bedlines) == 0:
        raise UsageError(
            f"Primer chrom names ({', '.join(bed_chrom_names)}) not found in MSA ({', '.join(seqdict.keys())})"
        )

    kmers_names = [bedline.primername for bedline in bedlines]

    # Create mapping array
    # Find index of primary ref
    mapping_index = [i for i, (k, v) in enumerate(seqdict.items()) if k in primary_ref][
        0
    ]

    mapping_array, array = create_mapping(array, mapping_index)
    ref_index_to_msa_dict = ref_index_to_msa(mapping_array)

    # Group Primers by basename
    basename_to_line: dict[str, set[BedLine]] = {
        "_".join(name.split("_")[:-1]): set() for name in kmers_names
    }
    for bedline in bedlines:
        basename = "_".join(bedline.primername.split("_")[:-1])
        basename_to_line[basename].add(bedline)

    basename_to_index = {bn: i for i, bn in enumerate(basename_to_line.keys())}

    seq_to_primername = {line.sequence: line.primername for line in bedlines}

    # Create the scoremap
    scoremap = np.empty((array.shape[0], len(basename_to_line)))
    scoremap.fill(None)
    textmap = np.empty((array.shape[0], len(basename_to_line)), dtype="str")
    textmap.fill("None")
    textmap = textmap.tolist()

    # get FPrimer sequences for each basename
    for bn, lines in basename_to_line.items():
        # Get primer size
        primer_len_max = max(len(line.sequence) for line in lines)

        # Set the direction
        if "LEFT" in bn:
            forward = True
            primer_end = list(lines)[0].end
            # Check for the end on a gap edge case and fix it
            if check_for_end_on_gap(ref_index_to_msa_dict, primer_end):
                msa_index = fix_end_on_gap(ref_index_to_msa_dict, primer_end)
            else:
                msa_index = ref_index_to_msa_dict[list(lines)[0].end]
        else:
            forward = False
            msa_index = ref_index_to_msa_dict[list(lines)[0].start]

        # Get the primer sequences
        msa_data = get_primers_from_msa(array, msa_index, forward, primer_len_max)

        # Get the score for each genome
        primer_seqs = {line.sequence for line in lines}

        for genome_index, genome_seq in msa_data.items():
            # Caused by gaps in the msa
            if genome_seq is None:
                if forward:
                    slice = array[genome_index, msa_index - primer_len_max : msa_index]
                    slice[slice == ""] = "-"
                    genome_seq = "".join(slice)
                else:
                    slice = array[genome_index, msa_index : msa_index + primer_len_max]
                    slice[slice == ""] = "-"
                    genome_seq = reverse_complement("".join(slice))

                textmap[genome_index][basename_to_index[bn]] = PlotlyText(
                    primer_seq=[x for x in primer_seqs][0],
                    genome_seq="".join(slice),
                    primer_name=bn,
                ).format_str()
                continue
            # Quick check for exact match
            if genome_seq in primer_seqs:
                scoremap[genome_index, basename_to_index[bn]] = 0
                primer_seq = "".join(primer_seqs.intersection({genome_seq}))
                textmap[genome_index][basename_to_index[bn]] = PlotlyText(
                    primer_seq=primer_seq,
                    genome_seq=genome_seq,
                    primer_name=seq_to_primername.get(primer_seq, "Unknown"),
                ).format_str()
                continue
            # Calculate the hamming distance between all
            seq_to_scores: dict[str, int] = {}
            for primer_seq in primer_seqs:
                seq_to_scores[primer_seq] = calc_primer_hamming(primer_seq, genome_seq)
            scoremap[genome_index, basename_to_index[bn]] = min(
                seq_to_scores.values(),  # type: ignore
            )
            primer_seq = "".join(
                [
                    k
                    for k, v in seq_to_scores.items()
                    if v == scoremap[genome_index, basename_to_index[bn]]
                ][0]
            )
            textmap[genome_index][basename_to_index[bn]] = PlotlyText(
                genome_seq=genome_seq,
                primer_seq=primer_seq,
                primer_name=seq_to_primername.get(primer_seq, "Unknown"),
            ).format_str()

    # Hovertemplate string
    if include_seqs:
        hovertemplatestr = "%{text}<br>" + "<b>Mismatches: %{z}</b><br>"
    else:
        hovertemplatestr = ""

    # Create the heatmap
    fig = go.Figure(
        data=go.Heatmap(
            z=scoremap,
            x=list(basename_to_line.keys()),
            y=[x for x in seqdict.keys()],
            colorscale="Viridis",
            text=textmap if include_seqs else None,  # only show text if not minimal
            hovertemplate=hovertemplatestr,
            xgap=0.1,
            ygap=0.1,
            name="Primer Mismatches",
        )
    )
    fig.update_layout(
        font=dict(family="Courier New, monospace"),
        hoverlabel=dict(font_family="Courier New, monospace"),
        title_text=f"Primer Mismatches: {list(primary_ref)[0]}",
    )
    fig.update_yaxes(autorange="reversed")

    # Remove unnecessary plot elements
    fig.update_layout(
        modebar_remove=[
            "select2d",
            "lasso2d",
            "select",
            "autoScale2d",
            "zoom",
            "toImage",
        ]
    )

    return fig.to_html(
        include_plotlyjs=True if offline_plots else "cdn", full_html=False
    )


def bedfile_plot_html(
    bedfile: pathlib.Path, ref_name: str, ref_seq: str, offline_plots: bool = False
) -> str:
    """
    Create a plotly heatmap from a bedfile.
    """
    # Read in the bedfile
    primerpairs, _header = read_in_bedprimerpairs(bedfile)

    # Filter primerpairs for the reference genome
    wanted_primerspairs = [pp for pp in primerpairs if pp.chrom_name == ref_name]
    if len(wanted_primerspairs) == 0:
        raise ValueError(f"No primers found for {ref_name}")

    # Calculate the GC data
    ref_array = np.array(
        [ref_seq.upper()],
        dtype="U1",
        ndmin=2,
    )
    gc = calc_gc(ref_array)

    # Create the plot
    fig = make_subplots(
        cols=1,
        rows=2,
        shared_xaxes=True,
        row_heights=[1, 0.5],
        specs=[
            [{"secondary_y": False}],
            [{"secondary_y": False}],
        ],
    )

    # Add the GC data
    fig.add_trace(
        go.Scattergl(
            x=[x[0] for x in gc],
            y=[x[1] for x in gc],
            mode="lines",
            name="GC Prop",
            line=dict(color="#005c68", width=2),
            fill="tozeroy",
        ),
        row=2,
        col=1,
    )

    # Add the primer lines
    for pp in wanted_primerspairs:
        print(f"Adding primer: {pp._primername}")
        fig.add_shape(
            type="rect",
            y0=pp.pool + 1 - 0.05,
            y1=pp.pool + 1 + 0.05,
            x0=pp.fprimer.region()[0],
            x1=pp.fprimer.region()[1],
            fillcolor="LightSalmon",
            line=dict(color="darksalmon", width=2),
            row=1,
            col=1,
        )
        fig.add_shape(
            type="rect",
            y0=pp.pool + 1 - 0.05,
            y1=pp.pool + 1 + 0.05,
            x0=pp.rprimer.region()[0],
            x1=pp.rprimer.region()[1],
            fillcolor="LightSalmon",
            line=dict(color="darksalmon", width=2),
            row=1,
            col=1,
        )
        # Handle circular genomes
        is_circular = pp.fprimer.region()[0] > pp.rprimer.region()[1]
        fig.add_shape(
            type="line",
            y0=pp.pool + 1,
            y1=pp.pool + 1,
            x0=pp.fprimer.region()[1],
            x1=pp.rprimer.region()[0] if not is_circular else len(ref_seq),
            line=dict(color="LightSeaGreen", width=5),
            row=1,
            col=1,
        )
        if is_circular:
            fig.add_shape(
                type="line",
                y0=pp.pool + 1,
                y1=pp.pool + 1,
                x0=0,
                x1=pp.rprimer.region()[0],
                line=dict(color="LightSeaGreen", width=5),
                row=1,
                col=1,
            )

    fig.update_xaxes(
        showline=True,
        mirror=True,
        ticks="outside",
        linewidth=2,
        linecolor="black",
        tickformat=",d",
        title_font=dict(size=18, family="Arial", color="Black"),
        range=[0, len(ref_seq)],
        title="",  # Blank title for all x-axes
    )
    fig.update_yaxes(
        showline=True,
        mirror=True,
        ticks="outside",
        linewidth=2,
        linecolor="black",
        fixedrange=True,
        title_font=dict(size=18, family="Arial", color="Black"),
    )
    # Update the top plot
    pools = sorted({x.pool + 1 for x in wanted_primerspairs})
    fig.update_yaxes(
        range=[pools[0] - 0.5, pools[-1] + 0.5],
        title="pool",
        tickmode="array",
        tickvals=pools,
        row=1,
        col=1,
    )
    fig.update_yaxes(
        range=[-0.1, 1.1],
        title="GC%",
        tickmode="array",
        tickvals=[0, 0.25, 0.5, 0.75, 1],
        ticktext=[0, 25, 50, 75, 100],
        row=2,
        col=1,
    )

    # Remove unnecessary plot elements
    fig.update_layout(
        modebar_remove=[
            "select2d",
            "lasso2d",
            "select",
            "autoScale2d",
            "zoom",
            "toImage",
        ]
    )
    fig.update_layout(height=400, title_text=ref_name, showlegend=False)

    # Write a html version of the plot
    return fig.to_html(
        include_plotlyjs=True if offline_plots else "cdn", full_html=False
    )



from enum import Enum
from itertools import groupby
from typing import Iterable

from primer3 import calc_hairpin as p3_calc_hairpin
from primer3 import calc_tm as p3_calc_tm

from primalscheme3.core.config import Config


class THERMORESULT(Enum):
    # THERMORESULT.value == 0 is a pass
    PASS = 0
    HIGH_GC = 1
    LOW_GC = 2
    HIGH_TM = 3
    LOW_TM = 4
    MAX_HOMOPOLY = 5
    HAIRPIN = 6


def calc_tm(kmer_seq, mv_conc, dv_conc, dntp_conc, dna_conc) -> float:
    """Return Tm for the kmer sequence."""
    return p3_calc_tm(
        kmer_seq,
        mv_conc=mv_conc,
        dv_conc=dv_conc,
        dntp_conc=dntp_conc,
        dna_conc=dna_conc,
    )


def calc_hairpin_tm(seq: str, mv_conc, dv_conc, dntp_conc, dna_conc) -> float:
    """
    Calculate the hairpin formation thermodynamics of a DNA sequence.
    Returns tm.
    """
    return p3_calc_hairpin(
        seq,
        mv_conc=mv_conc,
        dv_conc=dv_conc,
        dntp_conc=dntp_conc,
        dna_conc=dna_conc,
    ).tm


def calc_hairpin_struct(seq: str, mv_conc, dv_conc, dntp_conc, dna_conc) -> float:
    """
    Calculate the hairpin formation thermodynamics of a DNA sequence.
    Returns tm.
    """
    return p3_calc_hairpin(
        seq,
        mv_conc=mv_conc,
        dv_conc=dv_conc,
        dntp_conc=dntp_conc,
        dna_conc=dna_conc,
        output_structure=True,
    ).ascii_structure_lines


def forms_hairpin(seqs: Iterable[str], config: Config) -> bool:
    """
    Given a iterable of strings it will check the hairpin tm of all seqs
    If any form hairpins it will return True
    If all clear it will return False
    """
    for seq in seqs:
        if (
            calc_hairpin_tm(
                seq,
                mv_conc=config.mv_conc,
                dv_conc=config.dv_conc,
                dntp_conc=config.dntp_conc,
                dna_conc=config.dna_conc,
            )
            > config.primer_hairpin_th_max
        ):
            return True
    return False


def gc(kmer_seq: str) -> float:
    return round(100.0 * (kmer_seq.count("G") + kmer_seq.count("C")) / len(kmer_seq), 1)


def max_homo(kmer_seq) -> int:
    """Return max homopolymer length for the kmer sequence."""
    if not kmer_seq:
        return 0
    return max(sum(1 for _ in group) for _, group in groupby(kmer_seq))


def thermo_check(kmer_seq: str, config: Config) -> THERMORESULT:
    """Are all kmer thermo values below threshold?.

    Evaluation order.
    GC CHECK
    TM CHECK
    HOMOPOLY CHECK
    HAIRPIN
    PASS

    Args:
        kmer_seq (str): The kmer sequence to be checked.
        cfg (dict): The configuration dictionary containing threshold values.

    Returns:
        THERMORESULT: The result of the thermo checks.
    """
    # Check for gc in range
    kmer_gc = gc(kmer_seq)
    if kmer_gc > config.primer_gc_max:
        return THERMORESULT.HIGH_GC
    elif kmer_gc < config.primer_gc_min:
        return THERMORESULT.LOW_GC

    # Check for tm in range
    kmer_tm = calc_tm(
        kmer_seq,
        mv_conc=config.mv_conc,
        dv_conc=config.dv_conc,
        dna_conc=config.dna_conc,
        dntp_conc=config.dntp_conc,
    )
    if kmer_tm > config.primer_tm_max:
        return THERMORESULT.HIGH_TM
    elif kmer_tm < config.primer_tm_min:
        return THERMORESULT.LOW_TM

    # Check for maxhomopolymer
    if max_homo(kmer_seq) > config.primer_homopolymer_max:
        return THERMORESULT.MAX_HOMOPOLY

    # Check for hairpin
    if (
        calc_hairpin_tm(
            kmer_seq,
            mv_conc=config.mv_conc,
            dv_conc=config.dv_conc,
            dntp_conc=config.dntp_conc,
            dna_conc=config.dna_conc,
        )
        > config.primer_hairpin_th_max
    ):
        return THERMORESULT.HAIRPIN

    return THERMORESULT.PASS


def thermo_check_all_kmers(
    kmers: Iterable[str], config: Config
) -> dict[str, THERMORESULT]:
    return {kmer: thermo_check(kmer, config) for kmer in kmers}


def thermo_check_kmers(kmers: Iterable[str], config: Config) -> THERMORESULT:
    """
    Will call thermo_check on each kmer sequence in the kmers list
    Will stop evaluating on first error

    Args:
        kmers (Iterable[str]): A list of kmer sequences to be evaluated.
        cfg (dict): A dictionary containing configuration settings.

    Returns:
        THERMORESULT: The result of the thermo checks. THERMORESULT.PASS if all kmers pass the checks, otherwise the first encountered error.

    """
    for kmer in kmers:
        result = thermo_check(kmer, config)
        if result == THERMORESULT.PASS:
            continue
        else:
            return result

    return THERMORESULT.PASS



import dbm.ndbm
from typing import Iterable

from Bio import SeqIO

# Module imports
from primalscheme3.core.seq_functions import expand_ambs, reverse_complement

MUTATIONS = {
    "A": "CGT",
    "C": "AGT",
    "G": "CAT",
    "T": "CGA",
}


class MatchDB:
    """
    Match encoding
    delim between matches:  b'*'
    delim between values:   b';'
    """

    def __init__(self, path, msas_paths: list[str], kmer_size: int) -> None:
        self.db = dbm.ndbm.open(str(path), "n")

        # Read in and digest each MSA
        for msa_index, msa_path in enumerate(msas_paths):
            records = SeqIO.parse(msa_path, "fasta")
            # For each sequence in the MSA
            for record in records:
                self._digest_kmers_into_db(
                    str(record.seq.upper()), kmer_size, msa_index
                )

    def _write_unique(self, sequence, match: tuple[int, int]):
        """This will only write unqiue values to the db"""
        match_bstr = f"{match[0]};{match[1]}".encode()

        if db_bstr := self.db.get(sequence):
            bmatches = {*db_bstr.split(b"*")}

            # If the new match bstr is not already in db
            if match_bstr not in bmatches:
                self.db[sequence] = db_bstr + b"*" + match_bstr

        else:
            self.db[sequence] = match_bstr

    def read_matches(self, sequence) -> list[list]:
        parsed_matches = []
        if db_string := self.db.get(sequence):
            matches = [*db_string.split(b"*")]

            for match in matches:
                d = match.split(b";")
                parsed_matches.append([int(d[0]), int(d[1])])

        return parsed_matches

    def find_match(self, sequence) -> list[list]:
        """
        Find all matches for a single sequence.

        :param sequence: The sequence to search for.
        :return: A list of matches, where each match is a list containing sequence details and orientation indicator.
        """
        matches = []

        # If the sequence is found
        if db_f_matches := self.read_matches(sequence):
            for match in db_f_matches:
                matches.append(match + ["+"])
        if db_r_matches := self.read_matches(reverse_complement(sequence)):
            for match in db_r_matches:
                matches.append(match + ["-"])

        return matches

    def find_matches(self, seqs: Iterable[str], fuzzy: bool = False) -> set[tuple]:
        """
        Find all matches for a collection of sequences.

        :param seqs: An iterable of sequences to search for.
        :param fuzzy: If True, consider fuzzy matches with single mismatches.
        :return: A set of matches, each represented as a tuple.
        """
        if fuzzy:
            search_seqs = {
                fseq
                for fseq in (generate_single_mismatches(seq) for seq in seqs)
                for fseq in fseq
            }
        else:
            search_seqs = seqs

        # Find all matches
        matches = {
            tuple(m) for m in (self.find_match(st) for st in search_seqs) for m in m
        }
        return matches

    def find_fkmer(
        self,
        fkmer,
        fuzzy: bool,
        kmersize: int,
        msaindex: int,
        remove_expected: bool,
    ):
        """
        Find matches for the given FKmer.

        :param fkmer: The FKmer to search for.
        :param fuzzy: If True, consider fuzzy matches with single mismatches.
        :param kmersize: The size of the kmer.
        :param msaindex: The index of the MSA.
        :param remove_expected: If True, remove expected matches.
        :return: A set of unexpected matches for the FKmer.
        """
        kmer_seqs = {x[-kmersize:] for x in fkmer.seqs}
        matches = self.find_matches(kmer_seqs, fuzzy)

        # Filter out expected matches
        if remove_expected:
            return {
                match
                for match in matches
                if match[1] != fkmer.end - kmersize and match[0] == msaindex
            }
        else:
            return matches

    def find_rkmer(
        self,
        rkmer,
        fuzzy: bool,
        kmersize: int,
        msaindex: int,
        remove_expected: bool,
    ):
        """
        Find unexpected matches for the given RKmer.

        :param rkmer: The RKmer to search for.
        :param fuzzy: If True, consider fuzzy matches with single mismatches.
        :param kmersize: The size of the kmer.
        :param msaindex: The index of the MSA.
        :param remove_expected: If True, remove expected matches.
        :return: A set of unexpected matches for the RKmer.
        """
        kmer_seqs = {x[:kmersize] for x in rkmer.seqs}
        matches = self.find_matches(kmer_seqs, fuzzy)

        # Filter out expected matches
        if remove_expected:
            return {
                match
                for match in matches
                if match[1] != rkmer.start and match[0] == msaindex
            }
        else:
            return matches

    def keys(self):
        return self.db.keys()

    def get(self, sequence, default=None):
        return self.db.get(sequence, default)

    def _digest_kmers_into_db(self, seq: str, kmer_size: int, msa_index):
        """
        Digests a sequence into kmers and adds them to the db

        :param seq: The sequence to digest.
        :param kmer_size: The Kmer size for digestion.
        :param kmersize: The size of the kmer.
        :param msa_index: The index of the MSA.
        """

        for i in range(len(seq) + 1 - kmer_size):
            # Prevent a kmer starting on invalid base
            if seq[i] in {"", "-"}:
                continue

            kmer = "".join(seq[i : i + kmer_size]).replace("-", "")

            if len(kmer) != kmer_size:
                counter = 0

                # Keep walking right until the kmer is the correct size or walks out of index
                while counter + kmer_size + i < len(seq) and len(kmer) < kmer_size:
                    new_base = seq[i + kmer_size + counter]

                    # If the new base in valid
                    if new_base != "-" and new_base != "":
                        kmer += new_base

                    counter += 1

            # Guard Check Kmer is correct size, and doesn't contain N
            if len(kmer) != kmer_size or "N" in kmer:
                continue
            # Expand any ambiguous bases
            if exp_kmers := expand_ambs([kmer]):
                # Write each sequence into the db
                for exp_kmer in exp_kmers:
                    self._write_unique(exp_kmer, (msa_index, i))
            else:
                continue


def generate_single_mismatches(base_seq: str) -> set[str]:
    """
    Generates a set of sequences with all single-base mismatches for the given base sequence.

    :param base_seq: The input base sequence.
    :return: A set containing base sequences with single-base mismatches
    """
    return_seqs = set([base_seq])
    base_seq_list = [x for x in base_seq]  # Split the seq into bases
    for mut_index, base in enumerate(base_seq_list):
        # handle invalid bases
        if alt_bases := MUTATIONS.get(base):
            for alt_base in alt_bases:
                return_seqs.add(
                    "".join(
                        base_seq_list[0:mut_index]
                        + [alt_base]
                        + base_seq_list[mut_index + 1 :]
                    )
                )
        else:
            raise ValueError(f"Invalid base '{base}' in sequence '{base_seq_list}'")
    return return_seqs


def detect_new_products(
    new_matches: set[tuple],
    old_matches: set[tuple],
    product_size: int = 2000,
) -> bool:
    """
    Detects if adding the new matches will result in interactions with the old matches.

    :param new_matches: A set of new matched sequences represented as tuples containing
                        the match position, MSA index, and orientation indicator.
    :param old_matches: A set of old matched sequences used for comparison.
    :param product_size: The maximum allowable product size.
    :return: True if interactions between new and old matches are detected, False otherwise.
    """
    # Split the new matches in forward and reverse
    fmatches = set()
    rmatches = set()
    for newmatch in new_matches:
        if newmatch[2] == "+":
            fmatches.add(newmatch)
        elif newmatch[2] == "-":
            rmatches.add(newmatch)

    # Split the old matches in forward and reverse
    old_fmatches = set()
    old_rmatches = set()
    for oldmatch in old_matches:
        if oldmatch[2] == "+":
            old_fmatches.add(oldmatch)
        elif oldmatch[2] == "-":
            old_rmatches.add(oldmatch)

    # Check the new fmatches against the old rmatches
    for fmatch in fmatches:
        for old_rmatch in old_rmatches:
            # If from same msa
            if fmatch[0] == old_rmatch[0]:
                # If within product distance
                if 0 < old_rmatch[1] - fmatch[1] < product_size:
                    return True

    # Check the new rmatches against the old fmatches
    for rmatch in rmatches:
        for old_fmatch in old_fmatches:
            # If from same msa
            if rmatch[0] == old_fmatch[0]:
                # If within product distance
                if 0 < rmatch[1] - old_fmatch[1] < product_size:
                    return True

    return False


def detect_products(matches: set[tuple[int, int, str]], product_size=2000) -> bool:
    """
    Detect the presence of potential product formations based on matched sequences.

    :param matches: A set of matched sequences represented as tuples containing the match position,
                    MSA index, and orientation indicator.
    :param product_size: The maximum allowable product size.
    :return: True if potential product formations are detected, False otherwise.
    """

    # TODO write a hash function that colides when a product is formed.
    ## Will make O(N) rather than than O(N^2)

    # Split the mew matches in forward and reverse
    fmatches = set()
    rmatches = set()
    for match in matches:
        if match[2] == "+":
            fmatches.add(match)
        elif match[2] == "-":
            rmatches.add(match)

    # If all matches are in the same direction
    if not fmatches or not rmatches:
        return False

    for fmatch in fmatches:
        for rmatch in rmatches:
            # If from same msa
            if fmatch[0] == rmatch[0]:
                # If within product distance
                if 0 < rmatch[1] - fmatch[1] < product_size:
                    return True

    return False


## It need to find the reverse complement of the primer in the forward sequence.
# kmerrc = "GTACGTCGATAG"
# kmer = "CTATCGACGTAC"
# "ACGATCGACTATCGACGTACGACATCGGACAGCAGATGTCGTACGTGATAGCTGCATGGTACGTCGATAG"
# "TGCTAGCTGATAGCTGCATGCTGTAGCCTGTCGTCTACAGCATGCACTATCGACGTACCATGCAGCTATC"
